import torch
import numpy as np
from scipy import optimize

def contrastive_learning(log_q_noise, log_q_data,
                         basis_noise, basis_data,
                         options = {'disp': True,
                                    'gtol': 1e-5}
):

    """
    Contrastive learning coefficients

    Parameters
    ----------
    log_q_noise: 1-dimensional tensor
        the logrithm of probability density for noise data under the noise distribution
    log_q_data: 1-dimensional tensor
        the logrithm of probability density for target data under the noise distribution
    basis_noise: 2-dimensional tensor
        the design matrix contraining basis values of noise data for compute the logrithm of probablity density for the target distribution
    basis_data: 2-dimensional tensor
        the design matrix contraining basis values of target data for compute the logrithm of probablity density for the target distribution

    Returns
    -------
    alpha: 
    
    """
    
    assert(basis_noise.shape[-1] == basis_data.shape[-1])
    assert(len(log_q_noise) == basis_noise.shape[0])
    assert(len(log_q_data) == basis_data.shape[0])

    basis_size = basis_noise.shape[-1]
    alphas = torch.zeros(basis_size, dtype=torch.float64)
    F = torch.zeros(1, dtype=torch.float64)

    x_init = np.concatenate([alphas.data.numpy(), F])
    
    def compute_loss_and_grad(x):
        alphas = torch.tensor(x[0:basis_size], requires_grad = True)
        F = torch.tensor(x[-1], requires_grad = True)

        u_data = torch.matmul(basis_data, alphas)
        u_noise = torch.matmul(basis_noise, alphas)

        num_samples_p = basis_data.shape[0]
        num_samples_q = basis_noise.shape[0]

        nu = F.new_tensor([num_samples_q / num_samples_p])
        
        log_p_data = - (u_data - F) - torch.log(nu)
        log_p_noise = - (u_noise - F) - torch.log(nu)

        log_q = torch.cat([log_q_noise, log_q_data])
        log_p = torch.cat([log_p_noise, log_p_data])

        logit = log_p - log_q
        target = torch.cat([torch.zeros_like(log_q_noise),
                            torch.ones_like(log_q_data)])
        loss = torch.nn.functional.binary_cross_entropy_with_logits(
            logit, target)               
        loss.backward()
        
        grad = torch.cat([alphas.grad, F.grad[None]]).numpy()

        return loss.item(), grad

    loss, grad = compute_loss_and_grad(x_init)
    
    
    results = optimize.minimize(compute_loss_and_grad,
                                x_init,
                                jac=True,
                                method='L-BFGS-B',
                                options = options)
    x = results['x']
    
    # x, f, d = optimize.fmin_l_bfgs_b(compute_loss_and_grad,
    #                                  x_init,
    #                                  iprint = 1,
    #                                  pgtol = 1e-6,
    #                                  factr = 100)

    alphas = x[0:basis_size]
    F = x[-1]

    return torch.from_numpy(alphas), F

def contrastive_learning_numpy(log_q_noise, log_q_data,
                               basis_noise, basis_data):

    """
    Contrastive learning coefficients

    Parameters
    ----------
    log_q_noise: 1-dimensional array
        the logrithm of probability density for noise data under the noise distribution
    log_q_data: 1-dimensional array
        the logrithm of probability density for target data under the noise distribution
    basis_noise: 2-dimensional array
        the design matrix contraining basis values of noise data for compute the logrithm of probablity density for the target distribution
    basis_data: 2-dimensional array
        the design matrix contraining basis values of target data for compute the logrithm of probablity density for the target distribution

    Returns
    -------
    alpha: 
    
    """
    
    assert(basis_noise.shape[-1] == basis_data.shape[-1])
    assert(len(log_q_noise) == basis_noise.shape[0])
    assert(len(log_q_data) == basis_data.shape[0])

    basis_size = basis_noise.shape[-1]
    alphas = np.zeros(basis_size)
    F = np.zeros(1)
    
    x_init = np.concatenate([alphas, F])

    log_q = np.concatenate([log_q_noise, log_q_data])
    y = np.concatenate([np.zeros_like(log_q_noise),
                        np.ones_like(log_q_data)])

    basis = np.concatenate([basis_noise, basis_data])

    num_samples_p = basis_data.shape[0]
    num_samples_q = basis_noise.shape[0]

    log_nu = np.log(num_samples_q / float(num_samples_p))
    
    def compute_loss_and_grad(x):
        alphas = x[0:basis_size]
        F = x[-1]

        ## compute loss = -(y*h - np.log(1 + np.exp(h)))
        h = -(np.matmul(basis, alphas) - F) - log_q - log_nu
        loss = -(y*h - (np.maximum(h, 0) + np.log(1 + np.exp(-np.abs(h)))))
        loss = np.mean(loss, 0)

        ## compute gradients
        p = 1. / (1 + np.exp(-np.abs(h)))
        p[h < 0] = 1 - p[h < 0]

        grad_alphas = np.matmul(basis.T, y - p) / y.shape[0]
        grad_F = -np.mean(y - p, keepdims = True)
        
        grad = np.concatenate([grad_alphas, grad_F])

        return loss, grad

    loss, grad = compute_loss_and_grad(x_init)
    x, f, d = optimize.fmin_l_bfgs_b(compute_loss_and_grad,
                                     x_init,
                                     iprint = 1)

    alphas = x[0:basis_size]
    F = x[-1]

    return alphas, F



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray
import os

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [5e-7,
                     1e-6, 2e-6]
#protein_names = protein_names.iloc[[1,-2], 0].tolist()
protein_names = protein_names.iloc[:, 0].tolist()

psf = {}
traj_ref = {}
for name in protein_names:
    print(name)
    psf[name] = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
    traj_ref[name] = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf[name])
    traj_ref[name].save_xyz(f"./output/figures/structure_with_lowest_rmsd/{name}_reference.xyz")

print("******* CG *********")        
rmsd_cg = {}
weight_decay = 1e-6

for name in protein_names:
    traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.dcd",
                              psf[name], stride = 4)
    traj_cg = traj_cg[10000:]
    rmsd_cg[name] = mdtraj.rmsd(traj_cg, traj_ref[name])

    idx = rmsd_cg[name].argmin()
    os.makedirs(f"./output/figures/structure_with_lowest_rmsd/", exist_ok = True)
    traj_cg.superpose(traj_ref[name])
    
    traj_cg[idx].save_xyz(f"./output/figures/structure_with_lowest_rmsd/{name}.xyz")

    rmsd = mdtraj.rmsd(traj_cg[idx], traj_ref[name])
    print(weight_decay, name, rmsd)
    
exit()



import numpy as np
from scipy.interpolate import BSpline
from scipy.integrate import quad
import pickle
import mdtraj
from sys import exit
import os
import argparse
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from CLCG.utils.splines import *
from CLCG.utils.CL import *
import torch
torch.set_default_dtype(torch.double)
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
import pandas as pd
from collections import defaultdict

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)

# job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
# job_idx = 0
# name = protein_names.iloc[job_idx, 0]

aa_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv")
aa_names = list(aa_info['name'])
aa_names.sort()
aa_pairs = []
for i in range(len(aa_names)):
    for j in range(i, len(aa_names)):
        aa_pairs.append((aa_names[i], aa_names[j]))
num_aa_pairs = len(aa_pairs)
r_min_dict = defaultdict(list)

def sort_pair_name(pair):
    if pair[0] <= pair[1]:
        return pair
    else:
        tmp = pair[0]
        pair[0] = pair[1]
        pair[1] = tmp
        return pair

for name in list(protein_names.iloc[:,0]):
    print(name)
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    resnames = [res.name for res in psf.residues]

    traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf, stride = 1)
    indices = [(i,j) for i in range(psf.n_residues) for j in range(i+4, psf.n_residues)]
    r = mdtraj.compute_distances(traj_md, indices)
    r_min = r.min(0)

    for k in range(len(indices)):
        i,j = indices[k]
        pair = sort_pair_name([resnames[i], resnames[j]])
        r_min_dict[tuple(pair)].append(r_min[k])

for key in r_min_dict.keys():
    r_min_dict[key] = np.min(r_min_dict[key])
r_min_dict = dict(r_min_dict)

os.makedirs(f"./output/common", exist_ok = True)    
with open(f"./output/common/LJ_rmin.pkl", 'wb') as file_handle:
    pickle.dump(r_min_dict, file_handle)
    



import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm

name = "2JOF"

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

traj_cg_lj = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_False_weight_decay_{4e-7:.3E}.dcd", psf)
traj_cg_rmsd = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_True_weight_decay_{2e-7:.3E}.dcd", psf)
traj_cg_nn = mdtraj.load_dcd(f"./output/{name}/nnforce_NVT/full_rmsd_False_weight_decay_4e-07_weight_decay_{1e-3:.3E}.dcd", psf)

traj_reference = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

rmsd_md = mdtraj.rmsd(traj_md, traj_reference, 0)
rmsd_cg_lj = mdtraj.rmsd(traj_cg_lj, traj_reference, 0)
rmsd_cg_rmsd = mdtraj.rmsd(traj_cg_rmsd, traj_reference, 0)
rmsd_cg_nn = mdtraj.rmsd(traj_cg_nn, traj_reference, 0)

exit()

fig = plt.figure(np.random.randint(0, 1000), figsize = (6.4*4, 4.8))
fig.clf()
plt.subplot(1,4,1)
n_md, bins, _ = plt.hist(rmsd_md, bins = 26, density = True, alpha = 0.6, label = 'All atom', log = False, color = 'C1')
n_cg_lj, _, _ = plt.hist(rmsd_cg_lj, bins = bins, density = True, alpha = 0.6, label = 'CG (pairwise)', log = False, color = 'C2')
# plt.xlabel("RMSD (nm)")
# plt.ylabel("Probability density")
plt.ylim(0, 4.0)
plt.legend()

plt.subplot(1,4,2)
n_md, bins, _ = plt.hist(rmsd_md, bins = bins, density = True, alpha = 0.6, label = 'All atom', log = False, color = 'C1')
n_cg_rmsd, _, _ = plt.hist(rmsd_cg_rmsd, bins = bins, density = True, alpha = 0.6, label = 'CG (pairwise + mb-rmsd)', log = False, color = 'C0')
# plt.xlabel("RMSD (nm)")
# plt.ylabel("Probability density")
plt.ylim(0, 4.0)
plt.legend()

plt.subplot(1,4,3)
n_md, bins, _ = plt.hist(rmsd_md, bins = bins, density = True, alpha = 0.6, label = 'All atom', log = False, color = 'C1')
n_cg_nn, _, _ = plt.hist(rmsd_cg_nn, bins = bins, density = True, alpha = 0.6, label = 'CG (pairwise + mb-network)', log = False, color = 'k')
# plt.xlabel("RMSD (nm)")
# plt.ylabel("Probability density")
plt.ylim(0, 4.0)
plt.legend()

plt.subplot(1,4,4)
F_md = -np.log(n_md)
F_cg_lj = -np.log(n_cg_lj)
F_cg_rmsd = -np.log(n_cg_rmsd)
F_cg_nn = -np.log(n_cg_nn)

plt.plot(0.5*(bins[1:] + bins[0:-1]), F_md, linewidth = 2, label = "All atom", color = 'C1')
plt.plot(0.5*(bins[1:] + bins[0:-1]), F_cg_lj, linewidth = 2, label = "CG (pairwise)", color = 'C2')
plt.plot(0.5*(bins[1:] + bins[0:-1]), F_cg_rmsd, linewidth = 2, label = "CG (pairwise + mb-rmsd)", color = 'C0')
plt.plot(0.5*(bins[1:] + bins[0:-1]), F_cg_nn, linewidth = 2, label = "CG (pairwise + mb-network])", color = 'k')
# plt.xlabel("RMSD (nm)")
# plt.ylabel("Free energy (kT)")
#plt.xlim(0, 1.8)
plt.ylim(None, 11.1)
plt.legend()
#plt.tight_layout()
#n, bins = np.histogram(rmsd_md, bins = 30, density = True)

os.makedirs("./output/plots", exist_ok = True)
fig.savefig("./output/plots/2JOF_rmsd_hist.eps")
fig.savefig("./output/plots/2JOF_rmsd_hist.pdf")

exit()



from scipy.interpolate import BSpline
from scipy.integrate import quad
import numpy as np
import math
import matplotlib.pyplot as plt
import torch

def bs(x, knots, boundary_knots, degree = 3, intercept = False):
    """ Generate the B-spline basis matrix for a polynomial spline.

    This function mimick the function bs in R package splines

    Args:
        x (Tensor): Values at which basis functions are evaludated.
        knots (Tensor): Internal breakpoints that define the spline.
        boundary_knots (Tensor): Boundary points 
        degree (int, optional): The degree of the piecewise polynomial. 
            The default is 3 for cubic splines.
        intercept (bool, optional): If True, an intercept is included 
            in the basis. Default is False.
    
    Returns:
        design_matrix (Tensor): A tensor of dimension (len(x), df), 
            where df = len(knots) + degree if intercept = False, 
            df = len(knots) + degree + 1 if intercept = True.
    """

    knots = knots.numpy()
    boundary_knots = boundary_knots.numpy()
    x = x.numpy()
    
    knots = np.concatenate([knots, boundary_knots])
    knots.sort()

    augmented_knots = np.concatenate([np.array([boundary_knots[0] for i in range(degree + 1)]),
                                      knots,
                                      np.array([boundary_knots[1] for i in range(degree + 1)])])
    num_of_basis = len(augmented_knots) - 2*(degree + 1) + degree + 1

    spl_list = []
    for i in range(num_of_basis):
        coeff = np.zeros(num_of_basis)
        coeff[i] = 1.0
        spl = BSpline(augmented_knots, coeff, degree, extrapolate = False)
        spl_list.append(spl)

    design_matrix = np.array([spl(x) for spl in spl_list]).T

    ## if the intercept is Fales, drop the first basis term, which is often
    ## referred as the "intercept". Note that np.sum(design_matrix, -1) = 1.
    ## see https://cran.r-project.org/web/packages/crs/vignettes/spline_primer.pdf
    if intercept is False:
        design_matrix = design_matrix[:, 1:]

    design_matrix = torch.from_numpy(design_matrix)

    return design_matrix

def pbs(x, knots, boundary_knots = torch.tensor([-math.pi, math.pi]), degree = 3, intercept = False):
    """ Compute the design matrix of a periodic B-spline. 

    This function mimick the pbs function in R package pbs.

    Args:
        x (Tensor): Values at which basis functions are evaludated.
        knots (Tensor): Internal breakpoints that define the spline.
        boundary_knots (Tensor): Boundary points 
        degree (int, optional): The degree of the piecewise polynomial. 
            The default is 3 for cubic splines.
        intercept (bool, optional): If True, an intercept is included 
            in the basis. Default is False.

    Returns:
        design_matrix (Tensor): A tensor of dimension (len(x), df), 
            where df = len(knots) if intercept = False, 
            df = len(knots) + 1 if intercept = True.
    """

    knots = knots.numpy()
    boundary_knots = boundary_knots.numpy()
    x = x.numpy()
    
    knots = np.concatenate([knots, boundary_knots])
    knots.sort()

    augmented_knots = np.copy(knots)
    for i in range(degree):
        augmented_knots = np.append(augmented_knots, knots[-1] + knots[i+1] - knots[0])
    for i in range(degree):
        augmented_knots = np.insert(augmented_knots, 0, knots[0] - (knots[-1] - knots[-1-(i+1)]))

    num_of_basis = len(augmented_knots) - 2*(degree + 1) + degree + 1

    spl_list = []
    for i in range(num_of_basis):
        coeff = np.zeros(num_of_basis)
        coeff[i] = 1.0
        spl = BSpline(augmented_knots, coeff, degree, extrapolate = False)
        spl_list.append(spl)

    design_matrix = np.array([spl(x) for spl in spl_list]).T
    design_matrix_left = design_matrix[:, 0:degree]
    design_matrix_right = design_matrix[:, -degree:]
    design_matrix_middle = design_matrix[:, degree:-degree]
    design_matrix = np.concatenate([design_matrix_middle, design_matrix_left + design_matrix_right], axis = -1)

    ## if the intercept is Fales, drop the first basis term, which is often
    ## referred as the "intercept".
    ## see https://cran.r-project.org/web/packages/crs/vignettes/spline_primer.pdf
    if intercept is False:
        design_matrix = design_matrix[:, 1:]

    design_matrix = torch.from_numpy(design_matrix)
    
    return design_matrix

def bs_lj(r, r_min, r_max, num_of_basis, omega = False):
    ''' Compute the design matrix of a custimized B-spline 
    for Lennard-Jones type interaction.

    Args:
        r (Tensor): Distances at which basis functions are evaluated.
        r_min (float): A cutoff distance. 
            When r < r_min, the interaction becomes repulsive and 
            the basis function will provide a postive value. 
        r_max (float): A cutoff distance. 
            When r > r_max, all basis functions are zeros.
        num_of_basis (int): The number of basis.
        omega (bool): Integral of secondary derivatives.
            If True, the function will also return a matrix omega.
            Omega[i,j] = \int_{r_min}^{r_max} 
                         basis_i.derivative(2)*basis_j.derivative(2) dr. 
            This matrix is useful when fitting a smoothing splines
            by addding a penaly term controling the secondary 
            derivative of splines.

    Returns:
        design_matrix (Tensor): A matrix of dimension (len(r), num_of_basis).
        omega (Tensor): A matrix containing the integral of the splines' 
            second derivatives
    '''
    r = r.numpy()
    
    ## degree of spline    
    degree = 3

    ## knots of cubic spline
    t = np.linspace(r_min, r_max + (r_max - r_min), num_of_basis*2 + 3)

    ## number of basis
    n = len(t) - 2 + degree + 1

    ## preappend and append knots
    t = np.concatenate(
        (np.array([r_min for i in range(degree)]),
         t,
         np.array([r_max + (r_max - r_min) for i in range(degree)])
        ))

    spl_list = []
    for i in range(n):
        c = np.zeros(n)
        c[i] = 1.0
        spl_list.append(BSpline(t, c, degree, extrapolate = True))

    spl_list = spl_list[:-(n//2+2)]    
    spl_list = [spl_list[i] for i in range(len(spl_list)) if i != 1]    

    design_matrix = []
    for i in range(len(spl_list)):
        u = spl_list[i](r)
        if i != 0:
            u[r <= r_min] = 0.
        design_matrix.append(u)
    design_matrix = np.array(design_matrix).T

    if omega:
        omega = np.zeros((len(spl_list), len(spl_list)))
        for i in range(len(spl_list)):
            for j in range(i, len(spl_list)):
                spl_i = spl_list[i].derivative(2)
                spl_j = spl_list[j].derivative(2)
                omega[i,j] = quad(lambda x: spl_i(x)*spl_j(x), r_min, r_max, limit = 10_000)[0]
                omega[j,i] = omega[i,j]
                
        omega[0,:] = 0.0
        omega[:,0] = 0.0
        
        return torch.from_numpy(design_matrix), torch.from_numpy(omega)

    else:
        return torch.from_numpy(design_matrix)

def bs_rmsd(r, r_max, num_of_basis):
    ''' Compute the design matrix of a custimized B-spline 
    for a biasing potential on RMSD.

    Args:
        r (Tensor): Distances at which basis functions are evaluated.
        r_max (float): A cutoff distance. 
            When r > r_max, all basis functions are zeros.
        num_of_basis (int): The number of basis.
    Returns:
        design_matrix (Tensor): A matrix of dimension (len(r), num_of_basis).
    '''
    r = r.numpy()
    
    ## degree of spline    
    degree = 3

    ## knots of cubic spline
    r_min = 0.0
    t = np.linspace(r_min, r_max + (r_max - r_min), num_of_basis*2 + 2)

    ## number of basis
    n = len(t) - 2 + degree + 1

    ## preappend and append knots
    t = np.concatenate(
        (np.array([r_min for i in range(degree)]),
         t,
         np.array([r_max + (r_max - r_min) for i in range(degree)])
        ))

    spl_list = []
    for i in range(n):
        c = np.zeros(n)
        c[i] = 1.0
        spl_list.append(BSpline(t, c, degree, extrapolate = True))

    spl_list = spl_list[:-(n//2+2)]    

    design_matrix = []
    for i in range(len(spl_list)):
        u = spl_list[i](r)
        design_matrix.append(u)
    design_matrix = np.array(design_matrix).T

    return torch.from_numpy(design_matrix)
    
if __name__ == "__main__":
    ## testing functions bs and pbs
    knots = torch.linspace(start = -math.pi, end = math.pi, steps = 10)
    knots = knots[1:-1]
    boundary_knots = torch.tensor([-math.pi, math.pi])
    
    x = torch.linspace(start = -math.pi, end = math.pi, steps = 200)
    degree = 3

    design_matrix_bs = bs(x, knots, boundary_knots, degree).numpy()
    design_matrix_pbs = pbs(x, knots, boundary_knots, degree).numpy()
    
    fig, axes = plt.subplots()
    for j in range(design_matrix_bs.shape[-1]):
        plt.plot(x, design_matrix_bs[:,j], label = f"{j}", linewidth = 3)
    #plt.legend()
    plt.tight_layout()
    fig.savefig("./output/design_matrix_bs.pdf")
    plt.close()
    
    fig, axes = plt.subplots()
    for j in range(design_matrix_pbs.shape[-1]):
        plt.plot(x, design_matrix_pbs[:,j], label = f"{j}")
    plt.legend()
    plt.tight_layout()
    fig.savefig("./output/design_matrix_pbs.pdf")
    plt.close()
    
    ## test the function bs_lj
    r_min, r_max = 0.3, 2.0
    num_of_basis = 12

    r = torch.linspace(r_min - 0.05, r_max + 1.0, 1000)
    design_matrix, omega = bs_lj(r, r_min, r_max, num_of_basis, True)

    fig, axes = plt.subplots()
    for j in range(design_matrix.shape[-1]):
        plt.plot(r, design_matrix[:,j], label = f"{j}")

    t = np.linspace(r_min, r_max + (r_max - r_min), num_of_basis*2 + 3)
    for i in range(len(t)):
        if t[i] <= r_max:
            plt.axvline(t[i], linestyle = '--')
    plt.legend()
    plt.tight_layout()
    fig.savefig("./output/design_matrix_bs_lj.pdf")
    plt.close()    

    ## test the function bs_rmsd
    r_min, r_max = 0.0, 2.0
    num_of_basis = 12

    r = torch.linspace(0.0, r_max + 1.0, 1000)
    design_matrix = bs_rmsd(r, r_max, num_of_basis)

    fig, axes = plt.subplots()
    for j in range(design_matrix.shape[-1]):
        plt.plot(r, design_matrix[:,j], label = f"{j}")
        
    t = np.linspace(r_min, r_max + (r_max - r_min), num_of_basis*2 + 3)
    for i in range(len(t)):
        if t[i] <= r_max:
            plt.axvline(t[i], linestyle = '--')
    plt.legend()
    plt.tight_layout()
    fig.savefig("./output/design_matrix_bs_rmsd.pdf")
    plt.close()    
    
    
    



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=nnforce
#SBATCH --time=10:00:00
#SBATCH --partition=xeon-g6-volta
#SBATCH --gres=gpu:volta:1
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0
#SBATCH --mem=100G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/nnforce_%a.txt

import argparse
import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pickle
from scipy import optimize
from sys import exit
import mdtraj
import pandas as pd
import simtk.unit as unit
from FastMBAR import *
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
sys.path.append("./script")
from NNForce import *
import os
from functools import reduce

name = '2JOF'
weight_decay_list = [1e-3]

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = 0
weight_decay = weight_decay_list[job_idx]

print(f"weight_decay: {weight_decay:.3E}", flush = True)

full_include_rmsd = False
full_weight_decay = 4e-7

print(f"full rmsd: {full_include_rmsd}, weight_decay: {full_weight_decay:.3E}")

with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
traj_noise = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay:.3E}.dcd", psf)

ic_md, _ = coor_transformer.compute_internal_coordinate_from_xyz(torch.from_numpy(traj_md.xyz))
ic_noise, _ = coor_transformer.compute_internal_coordinate_from_xyz(torch.from_numpy(traj_noise.xyz))

angle_md = torch.cat([ic_md.reference_particle_3_angle[:, None], ic_md.angle], dim = 1).double()
angle_noise = torch.cat([ic_noise.reference_particle_3_angle[:, None], ic_noise.angle], dim = 1).double()
angle = torch.cat([angle_noise, angle_md])

dihedral_md = ic_md.dihedral.double()
dihedral_noise = ic_noise.dihedral.double()
dihedral = torch.cat([dihedral_noise, dihedral_md])

p_index_for_distances = torch.LongTensor([[i,j] for i in range(psf.n_atoms) for j in range(i+4, psf.n_atoms)])
distance_md = utils.functional.compute_distances(torch.from_numpy(traj_md.xyz),
                                                 p_index_for_distances)

distance_noise = utils.functional.compute_distances(torch.from_numpy(traj_noise.xyz),
                                                   p_index_for_distances)
distance = torch.cat([distance_noise, distance_md])

target = torch.cat([torch.zeros(len(ic_noise)), torch.ones(len(ic_md))])

## log_p
hidden_size = 32
torch.manual_seed(0)
nnforce_net = NNForceNet(angle.shape[-1],
                         dihedral.shape[-1],
                         distance.shape[-1],
                         hidden_size)

with torch.no_grad():
    nnforce_energy = nnforce_net(angle, dihedral, distance)
    
energy_md = nnforce_energy
energy_noise = torch.zeros_like(energy_md)

energy_matrix = torch.stack([energy_noise, energy_md]).numpy()
num_conf = np.array([traj_noise.n_frames, traj_md.n_frames])
fastmbar = FastMBAR(energy_matrix, num_conf, verbose = True)
F = fastmbar.F[-1]

## contrastive learning
if torch.cuda.is_available():
    device = torch.device("cuda")
else:
    device = torch.device("cpu")

nnforce_net = nnforce_net.to(device)
F = torch.tensor([F], device = device, requires_grad = True)

angle = angle.to(device)
dihedral = dihedral.to(device)
distance = distance.to(device)
energy_noise = energy_noise.to(device)
log_q = -energy_noise
target = target.to(device)

num_eval = [0]

params = list(nnforce_net.parameters()) + [F]

def gather_flat(params, grad = False):
    views = []
    for p in params:
        if grad:
            if p.grad is None:
                view =  p.new(p.numel()).zero_()
            else:
                view = p.grad.view(-1)
        else:
            view = p.data.view(-1)
        views.append(view)
    return torch.cat(views, 0)

def set_param(params, update):
    update = update.to(device)
    offset = 0
    for p in params:
        numel = p.numel()
        p.data.copy_(update[offset:offset + numel].view_as(p))
        offset += numel

x_init = gather_flat(params).detach().clone().cpu().numpy()
set_param(params, torch.from_numpy(x_init))

def compute_loss_and_grad(x, weight_decay):
    ## set all gradients to zero
    for p in params:
        if p.grad is not None:
            p.grad.zero_()

    ## set parameters for the neural network
    set_param(params, torch.from_numpy(x))

    ## compute loss and gradient
    nnforce_energy = nnforce_net(angle, dihedral, distance)
    energy = nnforce_energy
    log_p = F - energy
    
    logit = log_p - log_q
    loss = nn.functional.binary_cross_entropy_with_logits(
        logit, target, reduction = 'mean')
    for p in nnforce_net.parameters():
        loss = loss + weight_decay*torch.sum(p**2)
        
    loss.backward()
    loss = loss.cpu().item()
    grad = gather_flat(params, grad = True)
    return loss, grad.cpu().numpy()

loss, grad = compute_loss_and_grad(x_init, weight_decay = 0.0)
x, f, d = optimize.fmin_l_bfgs_b(compute_loss_and_grad,
                                 x_init,
                                 args = [weight_decay],
                                 iprint = 1,
                                 maxfun = 50000,
                                 maxiter = 50000)

set_param(params, torch.from_numpy(x))
loss_wo_penalty, _ = compute_loss_and_grad(x, weight_decay = 0.0)

torch.save({'hidden_size': hidden_size,
            'state_dict':nnforce_net.cpu().state_dict(),
            'p_index_for_distances': p_index_for_distances,                        
            'loss_wo_penalty': loss_wo_penalty,},
           f"./output/{name}/FF/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_NNForce_weight_decay_{weight_decay:.3E}_different_initial_parameters.pth")



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=get_IC
#SBATCH --time=00:10:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --mem=30G
#SBATCH --output=./slurm_output/get_IC_%a.txt

import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import math
import argparse
import os
import mdtraj
from sys import exit
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
import torch
torch.set_default_dtype(torch.float)
import pickle
import pandas as pd

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = 10
name = protein_names.iloc[job_idx, 0]

print(f"protein: {name}")

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

ref_particle_1 = psf.n_atoms//2
ref_particle_2 = ref_particle_1 - 1
ref_particle_3 = ref_particle_1 + 1

bonds = {}
for i in range(psf.n_atoms):
    bonds[i] = []
    if i - 1 >= 0:
        bonds[i].append(i-1)
    if i + 1 < psf.n_atoms:
        bonds[i].append(i+1)

coor_transformer = utils.CoordinateTransformer(
    bonds,
    ref_particle_1,
    ref_particle_2,
    ref_particle_3,
    dihedral_mode = 'linear'
)

os.makedirs(f"./output/{name}/", exist_ok = True)
with open(f"./output/{name}/coor_transformer.pkl", 'wb') as file_handle:
    pickle.dump(coor_transformer, file_handle)

traj = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf)

xyz = traj.xyz.astype(np.float64)
ic, ic_logabsdet = coor_transformer.compute_internal_coordinate_from_xyz(torch.from_numpy(xyz))

exit()

os.makedirs(f"./output/{name}/", exist_ok = True)
torch.save({'ic': ic, 'ic_logabsdet': ic_logabsdet},
           f"./output/{name}/ic_md.pt")

ic.plot(file_name = f"./output/{name}/ic_md_hist.pdf")



import os
import pickle
import pandas as pd
import numpy as np
import torch
from sys import exit

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_names = protein_names.iloc[:, 0].tolist()

elec_energy_fshift = {}
elec_energy_simple = {}
for name in protein_names:
    print(name)
    data = torch.load(f"./output/{name}/elec_energy_fshift.pt")
    elec_energy_fshift[name] = torch.cat([data['imus'], data['md']])

    data = torch.load(f"./output/{name}/elec_energy_simple.pt")
    elec_energy_simple[name] = torch.cat([data['imus'], data['md']])
    

exit()

rmsd_centers_and_k = {}
for name in protein_names:
    with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
        data = pickle.load(file_handle)
        rmsd_centers_and_k[name] = data
sizes = [rmsd_centers_and_k[name]['size'] for name in protein_names]
cumsum_sizes = np.cumsum(sizes)

for idx_name in range(1, 4):
    for rank in range(cumsum_sizes[idx_name -1], cumsum_sizes[idx_name]):
        name = protein_names[idx_name]
        new_rank = rank - cumsum_sizes[idx_name -1]
        
        old_name = f"./output/{name}/traj_imus/traj_{rank}.dcd"
        new_name = f"./output/{name}/traj_imus/traj_{new_rank}.dcd"

        os.rename(old_name, new_name)
        print(name, old_name, new_name)




#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=tica
#SBATCH --time=3:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --array=0-19
#SBATCH --mem=100G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/tica_%a.txt

import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm
import math
from sklearn.mixture import GaussianMixture
import pandas as pd

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_names = protein_names.iloc[:, 0].tolist()
weight_decay = 1e-6

os.makedirs('./output/figures/tica', exist_ok = True)

## CLN025
############################################################
name = 'CLN025'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## 2JOF
############################################################
name = '2JOF'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = -15, -15
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## 1FME
############################################################
name = '1FME'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## 2F4K
############################################################
name = '2F4K'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## GTT
############################################################
name = 'GTT'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 5000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## NTL9
############################################################
name = 'NTL9'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 10000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 30
levels = np.linspace(F_min, F_max, 10)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## 2WAV
############################################################
name = '2WAV'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## PRB
############################################################
name = 'PRB'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## UVF
############################################################
name = 'UVF'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## NuG2
############################################################
name = 'NuG2'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## A3D
############################################################
name = 'A3D'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 10
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

## lambda
############################################################
name = 'lambda'
print(name)
psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)
index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_cg = mdtraj.compute_distances(traj_cg, index)
tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
x_cg = tica.get_output()[0]
x1_min, x2_min = x_cg.min(0)
x1_max, x2_max = x_cg.max(0)

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_cg = gaussian_kde(x_cg.T)
logp_grid_cg = kde_cg.logpdf(x_grid.T)
F_grid_cg = -logp_grid_cg.reshape((size, size))
F_min = math.floor(F_grid_cg.min())
F_max = F_min + 30
levels = np.linspace(F_min, F_max, 20)
fig = plt.figure()
fig.clf()
plt.contourf(F_grid_cg,
             levels = levels,
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
plt.colorbar()
plt.savefig(f'./output/figures/tica/tica_free_energy_{name}.eps')

exit()



for name in protein_names:
    print(name)
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", top = psf)
    traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_DH_2_ss_type_simple_weight_decay_1.000E-06.dcd", top = psf)

    index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
    distances_md = mdtraj.compute_distances(traj_md, index)
    distances_cg = mdtraj.compute_distances(traj_cg, index)

    # tica = pyemma.coordinates.tica(distances_md, lag = 1000, dim = 2, kinetic_map = False)
    # x_md = tica.get_output()[0]
    # x_cg = tica.transform(distances_cg)
    
    tica = pyemma.coordinates.tica(distances_cg, lag = 1000, dim = 2, kinetic_map = False)
    x_cg = tica.get_output()[0]
    x1_min, x2_min = x_cg.min(0)
    x1_max, x2_max = x_cg.max(0)
    
    #x_md = tica.transform(distances_md)

    # x1_min, x2_min = np.concatenate([x_md, x_cg]).min(0)
    # x1_max, x2_max = np.concatenate([x_md, x_cg]).max(0)
    #x2_max = 8.5

    size = 20
    x1_grid = np.linspace(x1_min, x1_max, size)
    x2_grid = np.linspace(x2_min, x2_max, size)

    x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
    x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

    # kde_md = gaussian_kde(x_md.T)
    # logp_grid_md = kde_md.logpdf(x_grid.T)
    # F_grid_md = -logp_grid_md.reshape((size, size))

    kde_cg = gaussian_kde(x_cg.T)
    logp_grid_cg = kde_cg.logpdf(x_grid.T)
    F_grid_cg = -logp_grid_cg.reshape((size, size))

    # F_min = F_grid_md.min()
    # F_grid_cg = F_grid_cg - F_grid_cg.min() + F_min
    # F_max = F_min + 10

    #colorbar_ticks = [i for i in range(math.ceil(F_min), F_max, 2)]
    
    # plt.subplot(12,2,protein_names.index(name)*2+1)
    # plt.contourf(F_grid_md,
    #              levels = np.linspace(F_min, F_max, 20),
    #              extent = (x1_min, x1_max, x2_min, x2_max),
    #              cmap = cm.viridis_r)
    #plt.colorbar(ticks = colorbar_ticks)

    plt.subplot(12,2,protein_names.index(name)*2+2)
    plt.contourf(F_grid_cg,
                 levels = 20,
                 extent = (x1_min, x1_max, x2_min, x2_max),
                 cmap = cm.viridis_r)
    #plt.colorbar(ticks = colorbar_ticks)

    #plt.colorbar(ticks = colorbar_ticks)

    # plt.xlabel(r"$x_1$", fontsize = 24)
    # plt.ylabel(r"$x_2$", fontsize = 24)
    # plt.tick_params(which='both', bottom=False, top=False, right = False, left = False, labelbottom=False, labelleft=False)
    #plt.tight_layout()
    #axes.set_aspect('equal')

plt.savefig(f"./output/figures/tica/tica_free_energy.eps")

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=make_reference_structure
#SBATCH --time=01:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/make_reference_%a.txt

import mdtraj
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import argparse
import numpy as np
import scipy.cluster.hierarchy
from scipy.spatial.distance import squareform
import pandas as pd
import os
from sys import exit

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf, stride = 1)

N = 5_000
stride = int(traj.n_frames / N)
traj_sub = traj[::stride]
traj_sub = traj_sub[0:N]

distances = np.empty((traj_sub.n_frames, traj_sub.n_frames))
for i in range(traj_sub.n_frames):
    distances[i] = mdtraj.rmsd(traj_sub, traj_sub, i)    
    if (i + 1) % 200 == 0:
        print(i)

print('Max pairwise rmsd: %f nm' % np.max(distances))
reduced_distances = squareform(distances, checks=False)

linkage = scipy.cluster.hierarchy.linkage(reduced_distances, method='average')
n_cluster = 50
cluster_label = scipy.cluster.hierarchy.fcluster(linkage, t = n_cluster, criterion = 'maxclust')
cluster_size = []
for i in range(1, n_cluster+1):
    cluster_size.append(np.sum(cluster_label == i))

popular_cluster_label = np.argmax(cluster_size) + 1
flag = cluster_label == popular_cluster_label
traj_sub = traj_sub[flag]

dist = np.mean(distances[flag,:][:,flag], -1)
idx = np.argmin(dist)

os.makedirs(f"./output/{name}/", exist_ok = True)
traj_sub[idx].save_xyz(f"./output/{name}/reference_structure.xyz")




#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=CL
#SBATCH --time=10:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --mem=100G
#SBATCH --array=0-39
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/cl_%a.txt

import argparse
import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pickle
from scipy import optimize
from sys import exit
import mdtraj
import pandas as pd
import simtk.unit as unit
from FastMBAR import *
import sys
import MMFlow
from MMFlow import utils
import pyarrow.feather as feather
import os
import math
import torch.nn as nn
from functools import reduce
from scipy.sparse import coo_matrix, csr_matrix
import itertools
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('--rank', type = int)
parser.add_argument('--size', type = int)
args = parser.parse_args()

rank = args.rank
size = args.size
print(f"rank: {rank}, size: {size}", flush = True)

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)
protein_names = protein_names.iloc[0:2, 0].tolist()
flag_rmsd = [False, True]
# weight_decay_list = [1e-7, 5e-7,
#                      1e-6, 5e-6,
#                      1e-5, 5e-5,
#                      1e-4, 5e-4,
#                      1e-3, 5e-3]
weight_decay_list = [2e-7, 3e-7, 4e-7]

options = list(itertools.product(protein_names, flag_rmsd, weight_decay_list))

name, include_rmsd, weight_decay = options[rank]
print(f"name: {name}, rmsd: {include_rmsd}, weight_decay: {weight_decay}", flush = True)

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

class CL_Actor(torch.nn.Module):
    def __init__(self, name, include_rmsd = False):
        super(CL_Actor, self).__init__()

        self.include_rmsd = include_rmsd
        
        ic_imus = torch.load(f"./output/{name}/ic_imus.pt")['ic']
        ic_md = torch.load(f"./output/{name}/ic_md.pt")['ic']

        self.N_imus = len(ic_imus)
        self.N_md = len(ic_md)
        self.N = self.N_imus + self.N_md
        
        self.target = torch.cat([torch.zeros(self.N_imus), torch.ones(self.N_md)])
        
        bond_imus = torch.cat(
            (ic_imus.reference_particle_2_bond[:, None],
             ic_imus.reference_particle_3_bond[:, None],
             ic_imus.bond), dim = -1)
        bond_md = torch.cat(
            (ic_md.reference_particle_2_bond[:, None],
             ic_md.reference_particle_3_bond[:, None],
             ic_md.bond), dim = -1)

        bond = torch.cat([bond_imus, bond_md])
        del ic_imus, ic_md, bond_imus, bond_md

        data = torch.load(f'./output/{name}/basis_angle_and_dihedrals.pt')
        self.rp3angle = torch.cat(
            [data['imus']['reference_particle_3_angle'],
             data['md']['reference_particle_3_angle']])
        self.angle = torch.cat(
            [data['imus']['angle'],
             data['md']['angle']])
        self.dihedral = torch.cat(
            [data['imus']['dihedral'],
             data['md']['dihedral']])
        
        self.angle_grid = data['grid']['angle_basis']
        self.angle_min = data['grid']['angle_min']        
        self.angle_max = data['grid']['angle_max']
        
        self.dihedral_grid = data['grid']['dihedral_basis']
        self.dihedral_min = data['grid']['dihedral_min']        
        self.dihedral_max = data['grid']['dihedral_max']
        del data

        with open(f"./output/{name}/bonded_parameters_im.pkl", 'rb') as file_handle:
            bonded_parameters = pickle.load(file_handle)
            self.bonded_parameters = bonded_parameters
            
        b0 = torch.cat([torch.tensor([bonded_parameters['reference_particle_2_bond']['b0']]),
                        torch.tensor([bonded_parameters['reference_particle_3_bond']['b0']]),
                        bonded_parameters['bond']['b0']])
        kb = torch.cat([torch.tensor([bonded_parameters['reference_particle_2_bond']['kb']]),
                        torch.tensor([bonded_parameters['reference_particle_3_bond']['kb']]),
                        bonded_parameters['bond']['kb']])
        bond_energy = torch.sum(0.5*kb*(bond - b0)**2, -1)
        self.register_buffer('bond_energy', bond_energy, persistent = False)
        
        self.c3a = nn.Parameter(bonded_parameters['reference_particle_3_angle']['alphas'])
        self.ca = nn.Parameter(torch.cat([p['alphas'] for p in bonded_parameters['angle']]))
        self.cd = nn.Parameter(torch.cat([p['alphas'] for p in bonded_parameters['dihedral']]))

        data = torch.load(f"./output/{name}/basis_lj.pt")
        self.lj = torch.cat([data['imus'], data['md']])
        self.lj_grid = data['grid']['basis']
        self.lj_min = data['grid']['min']
        self.lj_max = data['grid']['max']

        lj_omega = data['grid']['omega']
        self.register_buffer('lj_omega', lj_omega, persistent = False)
        self.n_lj, self.nb_lj = lj_omega.shape[0], lj_omega.shape[1]
        self.clj = nn.Parameter(torch.zeros(self.lj.shape[-1]))        
        del data

        
        if self.include_rmsd and os.path.exists(f"./output/{name}/basis_rmsd.pt"):
            data = torch.load(f"./output/{name}/basis_rmsd.pt")
            rmsd_imus = torch.cat([d['imus'] for d in data], dim = -1)
            rmsd_md = torch.cat([d['md'] for d in data], dim = -1)
            self.rmsd = torch.cat([rmsd_imus, rmsd_md])
            self.rmsd_grid = torch.stack([d['grid']['basis'] for d in data])
            self.rmsd_max = torch.tensor([d['grid']['max'] for d in data])            
            self.rmsd_min = torch.tensor([d['grid']['min'] for d in data])

            self.n_rmsd = len(data)
            self.nb_rmsd = self.rmsd_grid.shape[-1]
            del data, rmsd_imus, rmsd_md
            self.crmsd = nn.Parameter(torch.zeros(self.rmsd.shape[-1]))
        else:
            self.rmsd = None
            self.rmsd_min = None
            self.rmsd_max = None            
            self.n_rmsd = 0
            self.nb_rmsd = 0
            self.crmsd = nn.Parameter(torch.tensor([]))

        data = torch.load(f"./output/{name}/elec_energy.pt")
        elec_energy = torch.cat([data['imus'], data['md']])
        self.register_buffer('elec_energy', elec_energy, persistent = False)
        del data

        data = torch.load(f"./output/{name}/log_q_imus.pt")
        self.log_q = torch.cat([data['imus'], data['md']])

        self._numel_cache = None

        ## initialize self.F
        with torch.no_grad():
            u_p = self.bond_energy + self.elec_energy + \
                  self.rp3angle.mv(self.c3a) + self.angle.mv(self.ca) + \
                  self.dihedral.mv(self.cd)
        u_p = u_p.numpy()
        u_q = -self.log_q.numpy()

        energy_matrix = np.vstack([u_q, u_p])
        num_conf = np.array([self.N_imus, self.N_md], dtype = np.float64)
        fastmbar = FastMBAR(energy_matrix, num_conf, verbose = True)
        
        self.F = nn.Parameter(torch.tensor([fastmbar.F[-1]]))
        
        # ## Because pytorch 1.10 has limited support on sparse csr matrix,
        # ## we use scipy.sparse.csr_matrix for now
        # self.rp3angle = self._torch_coo_to_scipy_csr(self.rp3angle)        
        # self.angle = self._torch_coo_to_scipy_csr(self.angle)
        # self.dihedral = self._torch_coo_to_scipy_csr(self.dihedral)
        # self.lj = self._torch_coo_to_scipy_csr(self.lj)                
    
    def _numel(self):
        if self._numel_cache is None:
            self._numel_cache = reduce(lambda total, p: total + p.numel(), self.parameters(), 0)
        return self._numel_cache
    
    def _gather_flat_grad(self):
        views = []
        for p in self.parameters():
            if p.grad is None:
                view =  p.new(p.numel()).zero_()
            else:
                view = p.grad.view(-1)
            views.append(view)
        return torch.cat(views, 0)

    def _gather_flat_param(self):
        views = []
        for p in self.parameters():
            view = p.data.view(-1)
            views.append(view)
        return torch.cat(views, 0)
    
    def _set_param(self, update):
        offset = 0
        for p in self.parameters():
            numel = p.numel()
            p.data.copy_(update[offset:offset + numel].view_as(p))
            offset += numel
        assert offset == self._numel()                

    def _torch_coo_to_scipy_csr(self, s):
        s = s.coalesce()
        row, col = s.indices().tolist()
        data = s.values().numpy()
        shape = list(s.shape)
        return coo_matrix((data, (row, col)), shape = shape).tocsr()

    def _scipy_csr_to_torch_coo(self, s):
        s = s.tocoo()
        s = torch.sparse_coo_tensor(np.array([s.row, s.col]), s.data, size = s.shape)
        return s

    def make_parameter_bounds(self):
        bounds = []
        for name, param in self.named_parameters():
            if name == 'clj':
                for i,j in itertools.product(range(self.n_lj), range(self.nb_lj)):
                    if j == 0:
                        bounds.append([5, None])
                    else:
                        bounds.append([None, None])
            else:
                bounds += [[None, None] for _ in range(param.numel())]
        assert(len(bounds) == self._numel())
        return bounds
            
    def compute_loss_and_grad(self, x, weight_decay, batch_size):
        if type(x) == np.ndarray:
            x = torch.from_numpy(x)
            
        device = self.ca.device
        x = x.to(device)
        
        self._set_param(x)

        n_batch = self.N // batch_size
        if self.N % batch_size != 0: n_batch += 1

        tot_loss = 0
        self.zero_grad()        
        for k in range(n_batch):
            u_bond = self.bond_energy[k*batch_size:(k+1)*batch_size]
            u_elec = self.elec_energy[k*batch_size:(k+1)*batch_size]

            # rp3angle = self._scipy_csr_to_torch_coo(self.rp3angle[k*batch_size:(k+1)*batch_size]).to(device)        
            # angle = self._scipy_csr_to_torch_coo(self.angle[k*batch_size:(k+1)*batch_size]).to(device)
            # dihedral = self._scipy_csr_to_torch_coo(self.dihedral[k*batch_size:(k+1)*batch_size]).to(device)
            # lj = self._scipy_csr_to_torch_coo(self.lj[k*batch_size:(k+1)*batch_size]).to(device)

            rp3angle = self.rp3angle[k*batch_size:(k+1)*batch_size].to(device)        
            angle = self.angle[k*batch_size:(k+1)*batch_size].to(device)
            dihedral = self.dihedral[k*batch_size:(k+1)*batch_size].to(device)
            lj = self.lj[k*batch_size:(k+1)*batch_size].to(device)
            
            u = u_bond + u_elec + rp3angle.mv(self.c3a) + angle.mv(self.ca) + \
                dihedral.mv(self.cd) + lj.mv(self.clj)

            if self.rmsd is not None:
                rmsd = self.rmsd[k*batch_size:(k+1)*batch_size].to(device)
                u = u + rmsd.mv(self.crmsd)
                
            log_p = - (u - self.F) + u.new_tensor([self.N_md/self.N_imus])
            log_q = self.log_q[k*batch_size:(k+1)*batch_size].to(device)
            target = self.target[k*batch_size:(k+1)*batch_size].to(device)

            logit = log_p - log_q
            loss = torch.nn.functional.binary_cross_entropy_with_logits(
                logit, target, reduction = 'sum')        
            loss.backward()
            tot_loss += loss.cpu().item()
            
        loss = tot_loss / self.N
        for p in self.parameters():
            if p.grad is not None:
                p.grad = p.grad / self.N

        clj = self.clj.view(self.n_lj, self.nb_lj)
        loss_wc = 0.5*weight_decay*torch.sum(clj[:,None,:].matmul(self.lj_omega).matmul(clj[:,:,None]))
        loss_wc.backward()
        
        loss += loss_wc.cpu().item()
        grad = self._gather_flat_grad()

        return loss, grad.cpu().numpy()

    def get_FF(self):
        self.bonded_parameters['reference_particle_3_angle']['U'] = torch.matmul(self.angle_grid,
                                                                                 self.c3a.cpu().data)
        nb_angle = self.angle_grid.shape[-1]                
        for i in range(len(self.bonded_parameters['angle'])):
            ca = self.ca.cpu().data[i*nb_angle:(i+1)*nb_angle]
            self.bonded_parameters['angle'][i]['U'] = torch.matmul(self.angle_grid, ca)
            
        nb_dihedral = self.dihedral_grid.shape[-1]        
        for i in range(len(self.bonded_parameters['dihedral'])):
            cd = self.cd.cpu().data[i*nb_dihedral:(i+1)*nb_dihedral]
            self.bonded_parameters['dihedral'][i]['U'] = torch.matmul(self.dihedral_grid, cd)

        clj = self.clj.cpu().data.view(self.n_lj, 1, self.nb_lj)
        ulj = torch.sum(self.lj_grid*clj, dim = -1)

        urmsd = None

        if self.rmsd is not None:
            crmsd = self.crmsd.cpu().data.view(self.n_rmsd, 1, self.nb_rmsd)
            urmsd = torch.sum(self.rmsd_grid*crmsd, dim = -1)
            
        FF = {'bonded': self.bonded_parameters,
              'lj': {'U': ulj, 'min': self.lj_min, 'max': self.lj_max},
              'rmsd': {'U': urmsd, 'min': self.rmsd_min, 'max': self.rmsd_max}}
        
        return FF

actor = CL_Actor(name, include_rmsd)
x_init = actor._gather_flat_param().numpy()
loss, grad = actor.compute_loss_and_grad(x_init, weight_decay, actor.N)
x, f, d = optimize.fmin_l_bfgs_b(actor.compute_loss_and_grad,
                                 x_init,
                                 args = [weight_decay, actor.N],
                                 bounds = actor.make_parameter_bounds(),
                                 iprint = 1,
                                 pgtol = 1e-5)

FF = actor.get_FF()
state_dict = actor.state_dict()
loss, grad = actor.compute_loss_and_grad(x, 0.0, actor.N)

os.makedirs(f"./output/{name}/FF", exist_ok = True)
torch.save({'FF': FF, 'state_dict': state_dict, 'loss': loss},
           f"./output/{name}/FF/FF_rmsd_{include_rmsd}_weight_decay_{weight_decay:.3E}.pt")

exit()



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray
import os

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
weight_decay_list = [1e-10, 1e-9, 1e-8,
                     1e-7, 1e-6, 1e-5,
                     1e-4, 1e-3]
protein_names = protein_names.iloc[:, 0].tolist()

name = 'A3D'
weight_decay = 1e-5

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)    
traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf, stride = 10)
traj_cg = mdtraj.load_dcd(f"./output/{name}/NVT/weight_decay_{weight_decay:.3E}.dcd", psf, stride = 50)

os.makedirs(f"./output/test", exist_ok = True)
traj_ref.save_pdb(f"./output/test/ref.pdb")
os.makedirs(f"./output/test/md/", exist_ok = True)

# width = len(str(len(traj_md)))
# for i in range(len(traj_md)):
#     folder_name = i // 1000
#     os.makedirs(f"./output/test/md/{folder_name:04}", exist_ok = True)
#     traj_md[i].save_pdb(f"./output/test/md/{folder_name:04}/md_{i:0{width}}.pdb")
#     if (i + 1) % 100 == 0:
#         print(i)

width = len(str(len(traj_cg)))
for i in range(len(traj_cg)):
    folder_name = i // 1000
    os.makedirs(f"./output/test/cg/{folder_name:04}", exist_ok = True)
    traj_cg[i].save_pdb(f"./output/test/cg/{folder_name:04}/md_{i:0{width}}.pdb")
    if (i + 1) % 100 == 0:
        print(i)
        
exit()

@ray.remote    
def get_rmsd_cg(name, weight_decay):
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
    traj_cg_nvt = mdtraj.load_dcd(f"./output/{name}/NVT/weight_decay_{weight_decay:.3E}.dcd", psf, stride = 10)
    rmsd_cg_nvt = mdtraj.rmsd(traj_cg_nvt, traj_ref)
    return rmsd_cg_nvt

#ray.shutdown()
ray.init(ignore_reinit_error = True, _temp_dir = "/home/gridsan/dingxq/tmp/ray")

rmsd_cg_nvt = ray.get([get_rmsd_cg.remote(name, weight_decay) for name in protein_names for weight_decay in weight_decay_list])
rmsd_cg_nvt_dict = {}
idx = 0
for name in protein_names:
    rmsd_cg_nvt_dict[name] = {}
    print(name)
    for weight_decay in weight_decay_list:
        rmsd_cg_nvt_dict[name][weight_decay] = rmsd_cg_nvt[idx]
        idx += 1

for weight_decay in weight_decay_list:
    print(weight_decay)
    
    ncols_hist = 4
    nrows_hist = 3
    fig_hist = plt.figure(figsize = (6.4*ncols_hist, 4.8*nrows_hist))
    fig_hist.suptitle(f"weight_decay: {weight_decay:.3E}")

    ncols_traj = 1
    nrows_traj = 12
    fig_traj = plt.figure(figsize = (6.4*4, 4.8*nrows_traj))
    fig_traj.suptitle(f"weight_decay: {weight_decay:.3E}")
    
    for name in protein_names:
        print(name)
        rmsd_md = rmsd_md_dict[name]
        rmsd_cg_nvt = rmsd_cg_nvt_dict[name][weight_decay]
        
        axes = fig_hist.add_subplot(nrows_hist, ncols_hist, protein_names.index(name) + 1)
        axes.hist(rmsd_md, bins = 40, alpha = 0.5, label = 'md', density = True)
        axes.hist(rmsd_cg_nvt, bins = 40, alpha = 0.5, label = 'cg_nvt', density = True)
        axes.legend()
        axes.set_title(f"{name}")

        axes = fig_traj.add_subplot(nrows_traj, ncols_traj, protein_names.index(name) + 1)
        axes.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
        axes.plot(rmsd_cg_nvt[::len(rmsd_cg_nvt)//10000], alpha = 0.5, label = 'cg_nvt')
        axes.legend()
        axes.set_title(f"{name}")
        
    pdf_hist.savefig(fig_hist)
    pdf_traj.savefig(fig_traj)    

pdf_hist.close()
pdf_traj.close()

exit()

plt.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
plt.plot(rmsd_cg_nvt[::len(rmsd_cg_nvt)//10000], alpha = 0.5, label = 'cg_nvt')
plt.legend()
plt.title(f"{name}, weight_decay: {weight_decay:.3E}")
pdf_traj.savefig()
plt.close()

exit()



import mdtraj
import argparse
import math
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
# weight_decay_list = [1e-8,
#                      1e-7, 2e-7, 5e-7,
#                      1e-6, 2e-6, 5e-6,
#                      1e-5]

weight_decay_list = [5e-7,
                     1e-6, 2e-6]

protein_names = protein_names.iloc[:, 0].tolist()

pdf_hist =  PdfPages(f"./output/plots/rmsd_full_hist_elec_type_{elec_type}_ss_type_{ss_type}.pdf")
pdf_traj =  PdfPages(f"./output/plots/rmsd_full_traj_elec_type_{elec_type}_ss_type_{ss_type}.pdf")

rmsd_md_dict = {}
traj_ref_dict = {}
for name in protein_names:
    print(name)
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
    traj_ref_dict[name] = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)    
    traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf, stride = 10)
    rmsd_md_dict[name] = mdtraj.rmsd(traj_md, traj_ref_dict[name])

@ray.remote    
def get_rmsd_cg(name, weight_decay):
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
    traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
    traj_cg_nvt = mdtraj.load_dcd(f"./output/{name}/NVT/elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.dcd",
                                  psf, stride = 10)
    rmsd_cg_nvt = mdtraj.rmsd(traj_cg_nvt, traj_ref)
    return rmsd_cg_nvt

#ray.shutdown()
ray.init(ignore_reinit_error = True, _temp_dir = "/home/gridsan/dingxq/tmp/ray")

rmsd_cg_nvt = ray.get([get_rmsd_cg.remote(name, weight_decay) for name in protein_names for weight_decay in weight_decay_list])
rmsd_cg_nvt_dict = {}
idx = 0
for name in protein_names:
    rmsd_cg_nvt_dict[name] = {}
    print(name)
    for weight_decay in weight_decay_list:
        rmsd_cg_nvt_dict[name][weight_decay] = rmsd_cg_nvt[idx]
        idx += 1

for weight_decay in weight_decay_list:
    print(weight_decay)
    
    ncols_hist = 4
    nrows_hist = 3
    fig_hist = plt.figure(figsize = (6.4*ncols_hist, 4.8*nrows_hist))
    fig_hist.suptitle(f"weight_decay: {weight_decay:.3E}")

    ncols_traj = 1
    nrows_traj = 12
    fig_traj = plt.figure(figsize = (6.4*4, 4.8*nrows_traj))
    fig_traj.suptitle(f"weight_decay: {weight_decay:.3E}")
    
    for name in protein_names:
        print(name)
        rmsd_md = rmsd_md_dict[name]
        rmsd_cg_nvt = rmsd_cg_nvt_dict[name][weight_decay]
        
        axes = fig_hist.add_subplot(nrows_hist, ncols_hist, protein_names.index(name) + 1)
        axes.hist(rmsd_md, bins = 40, alpha = 0.5, label = 'md', density = True)
        axes.hist(rmsd_cg_nvt, bins = 40, alpha = 0.5, label = 'cg_nvt', density = True)
        axes.legend()
        axes.set_title(f"{name}")

        axes = fig_traj.add_subplot(nrows_traj, ncols_traj, protein_names.index(name) + 1)
        axes.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
        axes.plot(rmsd_cg_nvt[::len(rmsd_cg_nvt)//10000], alpha = 0.5, label = 'cg_nvt')
        axes.legend()
        axes.set_title(f"{name}")
        
    pdf_hist.savefig(fig_hist)
    pdf_traj.savefig(fig_traj)    

pdf_hist.close()
pdf_traj.close()

exit()

plt.plot(rmsd_md[::len(rmsd_md)//10000], alpha = 0.5, label = 'md')
plt.plot(rmsd_cg_nvt[::len(rmsd_cg_nvt)//10000], alpha = 0.5, label = 'cg_nvt')
plt.legend()
plt.title(f"{name}, weight_decay: {weight_decay:.3E}")
pdf_traj.savefig()
plt.close()

exit()



import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'large'
mpl.rcParams['xtick.labelsize'] = 'large'
mpl.rcParams['ytick.labelsize'] = 'large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm

name = "2JOF"

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

traj_cg_lj = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_False_weight_decay_{4e-7:.3E}.dcd", psf)
traj_cg_rmsd = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_True_weight_decay_{2e-7:.3E}.dcd", psf)
traj_cg_nn = mdtraj.load_dcd(f"./output/{name}/nnforce_NVT/full_rmsd_False_weight_decay_4e-07_weight_decay_{1e-3:.3E}.dcd", psf)

traj_reference = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

rmsd_md = mdtraj.rmsd(traj_md, traj_reference, 0)
rmsd_cg_lj = mdtraj.rmsd(traj_cg_lj, traj_reference, 0)
rmsd_cg_rmsd = mdtraj.rmsd(traj_cg_rmsd, traj_reference, 0)
rmsd_cg_nn = mdtraj.rmsd(traj_cg_nn, traj_reference, 0)

rmsd_max = max(rmsd_md.max(), rmsd_cg_lj.max(), rmsd_cg_rmsd.max(), rmsd_cg_nn.max())

N = 3000
rmsd_md = rmsd_md[::len(rmsd_md)//N]
rmsd_cg_lj = rmsd_cg_lj[::len(rmsd_cg_lj)//N]
rmsd_cg_rmsd = rmsd_cg_rmsd[::len(rmsd_cg_rmsd)//N]
rmsd_cg_nn = rmsd_cg_nn[::len(rmsd_cg_nn)//N]

fig = plt.figure(figsize = (6.4*4, 4.8*2))
fig.clf()
plt.subplot(4,1,1)
plt.plot(np.linspace(0, 208_000, len(rmsd_md)), rmsd_md, '.', label = 'All atom', ms = 6)
plt.ylim(0, rmsd_max)
plt.xlabel('Simulation time (ns)')
plt.ylabel('RMSD (nm)')

plt.subplot(4,1,2)
plt.plot(np.linspace(0, 500, len(rmsd_cg_lj)), rmsd_cg_lj, '.', label = 'CG (pairwise)', ms = 5)
plt.ylim(0, rmsd_max)
plt.xlabel('Simulation time (ns)')
plt.ylabel('RMSD (nm)')

plt.subplot(4,1,3)
plt.plot(np.linspace(0, 500, len(rmsd_cg_rmsd)), rmsd_cg_rmsd, '.', label = 'CG (pairse + mb-rmsd)', ms = 5)
plt.ylim(0, rmsd_max)
plt.xlabel('Simulation time (ns)')
plt.ylabel('RMSD (nm)')

plt.subplot(4,1,4)
plt.plot(np.linspace(0, 500, len(rmsd_cg_nn)), rmsd_cg_nn, '.', label = 'CG (pairse + mb-network)', ms = 5)
plt.ylim(0, rmsd_max)
plt.xlabel('Simulation time (ns)')
plt.ylabel('RMSD (nm)')

#plt.legend()
plt.tight_layout()
fig.savefig(f"./output/plots/{name}_rmsd_traj.eps")

exit()



import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm
import pickle

name = '2JOF'

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

traj_cg_lj = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_False_weight_decay_{4e-7:.3E}.dcd", psf)
traj_cg_rmsd = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_True_weight_decay_{2e-7:.3E}.dcd", psf)
traj_cg_nn = mdtraj.load_dcd(f"./output/{name}/nnforce_NVT/full_rmsd_False_weight_decay_4e-07_weight_decay_{1e-3:.3E}.dcd", psf)

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
num_residues = psf.n_residues
indices = []
for i in range(num_residues):
    for j in range(i+6, num_residues):
        indices.append([i,j])

distances_md = mdtraj.compute_distances(traj_md, np.array(indices))
distances_cg_lj = mdtraj.compute_distances(traj_cg_lj, np.array(indices))
distances_cg_rmsd = mdtraj.compute_distances(traj_cg_rmsd, np.array(indices))
distances_cg_nn = mdtraj.compute_distances(traj_cg_nn, np.array(indices))

nrow = 15
ncol = 7

fig = plt.figure(np.random.randint(0, 1000), figsize = (6.4*ncol, 4.8*nrow))
fig.clf()

for j in range(len(indices)):
    print(j)
    ax = plt.subplot(nrow, ncol, j+1)
    idx1, idx2 = indices[j]
    plt.hist(distances_md[:, j], bins = 100, density = True, alpha = 0.5, label = 'All atom', range = [0.0, 2.5], log = False, color = 'C1')
    plt.hist(distances_cg_nn[:, j], bins = 100, density = True, alpha = 0.5, label = r'CG (pairwise + mb-nn)', range = [0.0, 2.5], log = False, color = 'k')
    plt.title(f"{idx1}-{idx2}")
#    plt.yscale('log')
    plt.legend()

# handles, labels = ax.get_legend_handles_labels()
# fig.legend(handles, labels, loc = (0.29, 0.13))

plt.tight_layout()
fig.savefig(f"./output/plots/{name}_distances_nn.pdf")
plt.close()

exit()

## load spline coefficients
with open(f"./output/LJ/{name}/FF_learned_with_{noise_model}_as_noise/weight_decay_{weight_decay:.3E}.pkl", 'rb') as file_handle:
    FF = pickle.load(file_handle)
    LJ_parameters = FF['LJ_parameters']

pair_indices = list(LJ_parameters.keys())
pair_indices.sort()

nrow = 6
ncol = 4

fig = plt.figure(np.random.randint(0, 1000), figsize = (6.4*ncol, 4.8*nrow))
fig.clf()
for k in range(len(pair_indices)):
    plt.subplot(nrow, ncol, k+1)
    pair = pair_indices[k]
    i,j = pair_indices[k]

    aa1, aa2 = resnames[i], resnames[j]
    if resnames[j] < resnames[i]:
        aa1, aa2 = resnames[j], resnames[i]
    
    r_min, r_max = LJ_parameters[pair]['r_min'], LJ_parameters[pair]['r_max']
    ulj = LJ_parameters[pair]['ulj']
    r = np.linspace(r_min , r_max, len(ulj))
    plt.plot(r, ulj, linewidth = 6.0)
    plt.xlim(0, r_max)
    plt.ylim(ulj.min()-1, 7)

    plt.title(f"{i}-{j}:{aa1}-{aa2}")
    print(k)
    
fig.savefig(f"./output/plots/{name}_lj.eps")

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=tica
#SBATCH --time=3:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --array=0-19
#SBATCH --mem=100G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/tica_%a.txt

import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'x-large'
mpl.rcParams['xtick.labelsize'] = 'x-large'
mpl.rcParams['ytick.labelsize'] = 'x-large'
import numpy as np
import torch
import pyemma
import mdtraj
import os
from scipy.stats import gaussian_kde
from sys import exit
from matplotlib import cm
import math
from sklearn.mixture import GaussianMixture

name = "2JOF"

weight_decay_list = [0.0,
                     1e-10, 5e-10,  
                     1e-9 , 5e-9 ,
                     1e-8 , 5e-8 ,
                     1e-7 , 5e-7 ,
                     1e-6 , 5e-6 ,
                     1e-5 , 5e-5 ,
                     1e-4 , 5e-4 ,
                     1e-3 , 5e-3 ,
                     1e-2 , 5e-2 ,
                     1e-1 ]

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
#weight_decay = weight_decay_list[job_idx]
weight_decay = 1e-3

# full_include_rmsd = False
# full_weight_decay = 4e-7

full_include_rmsd = False
full_weight_decay = 4e-7

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

traj_cg_lj = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_False_weight_decay_{4e-7:.3E}.dcd", psf)
traj_cg_rmsd = mdtraj.load_dcd(f"./output/{name}/NVT/rmsd_True_weight_decay_{2e-7:.3E}.dcd", psf)
traj_cg_nn = mdtraj.load_dcd(f"./output/{name}/nnforce_NVT/full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay}_weight_decay_{weight_decay:.3E}.dcd", psf)
traj_reference = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

index = [ (i,j) for i in range(psf.n_residues) for j in range(i+2, psf.n_residues)]
distances_md = mdtraj.compute_distances(traj_md, index)
distances_cg_lj = mdtraj.compute_distances(traj_cg_lj, index)
distances_cg_rmsd = mdtraj.compute_distances(traj_cg_rmsd, index)
distances_cg_nn = mdtraj.compute_distances(traj_cg_nn, index)

tica = pyemma.coordinates.tica(distances_md, lag = 1000, dim = 2, kinetic_map = False)
x_md = tica.get_output()[0]

idx_0 = x_md[:,0] < -1
idx_1 = (x_md[:,0] >= -1) & (x_md[:,1] < 5)
idx_2 = (x_md[:,0] >= 0) & (x_md[:,1] >= 8)

traj_md_0 = traj_md[idx_0]
traj_md_1 = traj_md[idx_1]
traj_md_2 = traj_md[idx_2]

traj_md_0.save_dcd(f'./output/plots/{name}_md_0.dcd')
traj_md_1.save_dcd(f'./output/plots/{name}_md_1.dcd')
traj_md_2.save_dcd(f'./output/plots/{name}_md_2.dcd')

exit()

# x = np.copy(x_md[x_md[:,1] < 5])
# gm = GaussianMixture(n_components = 2)
# gm.fit(x)

# x = np.copy(x_md[(x_md[:,1] > 6) & (x_md[:,0] > 0)])
# centers = np.concatenate([gm.means_, x.mean(0, keepdims = True)])
# print(centers)

# traj_md = traj_md.superpose(traj_reference, 0)
# for i in range(centers.shape[0]):
#     d = np.sqrt(np.sum((x_md - centers[i])**2, -1))
#     idx = d.argsort()
#     idx = idx[0:5]
#     traj_md[idx].save_dcd(f"./output/plots/{name}_center_{i}.dcd")

# exit()

x_cg_lj = tica.transform(distances_cg_lj)
x_cg_rmsd = tica.transform(distances_cg_rmsd)
x_cg_nn = tica.transform(distances_cg_nn)

x1_min, x2_min = np.concatenate([x_md, x_cg_lj, x_cg_rmsd, x_cg_nn]).min(0)
x1_max, x2_max = np.concatenate([x_md, x_cg_lj, x_cg_rmsd, x_cg_nn]).max(0)
#x2_max = 8.5

size = 20
x1_grid = np.linspace(x1_min, x1_max, size)
x2_grid = np.linspace(x2_min, x2_max, size)

x1_grid, x2_grid = np.meshgrid(x1_grid, x2_grid)
x_grid = np.stack([x1_grid.reshape(-1), x2_grid.reshape(-1)], 1)

kde_md = gaussian_kde(x_md.T)
logp_grid_md = kde_md.logpdf(x_grid.T)
F_grid_md = -logp_grid_md.reshape((size, size))

kde_cg_lj = gaussian_kde(x_cg_lj.T)
logp_grid_cg_lj = kde_cg_lj.logpdf(x_grid.T)
F_grid_cg_lj = -logp_grid_cg_lj.reshape((size, size))

kde_cg_rmsd = gaussian_kde(x_cg_rmsd.T)
logp_grid_cg_rmsd = kde_cg_rmsd.logpdf(x_grid.T)
F_grid_cg_rmsd = -logp_grid_cg_rmsd.reshape((size, size))

kde_cg_nn = gaussian_kde(x_cg_nn.T)
logp_grid_cg_nn = kde_cg_nn.logpdf(x_grid.T)
F_grid_cg_nn = -logp_grid_cg_nn.reshape((size, size))

F_min = F_grid_md.min()
F_grid_cg_lj = F_grid_cg_lj - F_grid_cg_lj.min() + F_min
F_grid_cg_rmsd = F_grid_cg_rmsd - F_grid_cg_rmsd.min() + F_min
F_max = 9

colorbar_ticks = [i for i in range(math.ceil(F_min), F_max, 2)]

fig = plt.figure(figsize = (6.4*5, 4.8))
fig.clf()

plt.subplot(1,5,1)
plt.contourf(F_grid_md,
             levels = np.linspace(F_min, F_max, 20),
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
#plt.colorbar(ticks = colorbar_ticks)

plt.subplot(1,5,2)
plt.contourf(F_grid_cg_lj,
             levels = np.linspace(F_min, F_max, 20),             
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)
#plt.colorbar(ticks = colorbar_ticks)

plt.subplot(1,5,3)
plt.contourf(F_grid_cg_rmsd,
             levels = np.linspace(F_min, F_max, 20),                          
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)

#plt.colorbar(ticks = colorbar_ticks)

plt.subplot(1,5,4)
plt.contourf(F_grid_cg_nn,
             levels = np.linspace(F_min, F_max, 20),                          
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)

plt.subplot(1,5,5)
plt.contourf(F_grid_cg_nn,
             levels = np.linspace(F_min, F_max, 20),                          
             extent = (x1_min, x1_max, x2_min, x2_max),
             cmap = cm.viridis_r)

plt.colorbar(ticks = colorbar_ticks)

# plt.xlabel(r"$x_1$", fontsize = 24)
# plt.ylabel(r"$x_2$", fontsize = 24)
# plt.tick_params(which='both', bottom=False, top=False, right = False, left = False, labelbottom=False, labelleft=False)
#plt.tight_layout()
#axes.set_aspect('equal')

plt.savefig(f"./output/plots/{name}_full_rmsd_{full_include_rmsd}_weight_decay_{full_weight_decay:.3E}_tica_density_{weight_decay:.3E}.eps")

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=bad_dist
#SBATCH --time=00:10:00
#SBATCH --partition=xeon-p8
#SBATCH --exclusive
#SBATCH --nodes=1
#SBATCH --array=0-1
#SBATCH --mem=30G
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/bad_dist_%a.txt

import mdtraj
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
from matplotlib.backends.backend_pdf import PdfPages
import argparse
import numpy as np
import scipy.cluster.hierarchy
from scipy.spatial.distance import squareform
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
import pickle
from sys import exit
import pandas as pd
import os
import torch

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)
    
data = torch.load(f"./output/{name}/ic_md.pt")
ic_md = data['ic']

traj_im = mdtraj.load_dcd(f"./output/{name}/traj_im/traj.dcd", psf)
xyz = torch.from_numpy(traj_im.xyz)
ic_im, _ = coor_transformer.compute_internal_coordinate_from_xyz(xyz)

bond_md = torch.cat([ic_md.reference_particle_2_bond[:, None],
                     ic_md.reference_particle_3_bond[:, None],
                     ic_md.bond], dim = -1).numpy()

bond_im = torch.cat([ic_im.reference_particle_2_bond[:, None],
                     ic_im.reference_particle_3_bond[:, None],
                     ic_im.bond], dim = -1).numpy()

angle_md = torch.cat([ic_md.reference_particle_3_angle[:, None],
                      ic_md.angle], dim = -1).numpy()

angle_im = torch.cat([ic_im.reference_particle_3_angle[:, None],
                      ic_im.angle], dim = -1).numpy()

dihedral_md = ic_md.dihedral.numpy()
dihedral_im = ic_im.dihedral.numpy()

num_columns = 4
num_rows = (ic_im.angle.shape[-1]*3 + 3) // num_columns + 1
fig = plt.figure(np.random.randint(1000), figsize = (6.4*num_columns, 4.8*num_rows))
fig.clf()

subplot_idx = 1
stride = 10

for j in range(bond_md.shape[-1]):
    plt.subplot(num_rows, num_columns, subplot_idx)    
    plt.hist(bond_md[::stride,j], bins = 30, density = True, label = f"bond_md", alpha = .5)
    plt.hist(bond_im[::stride,j], bins = 30, density = True, label = f"bond_im", alpha = .5)
    plt.title('bond')
    plt.legend()
    subplot_idx += 1
    print(subplot_idx)
    
for j in range(angle_md.shape[-1]):
    plt.subplot(num_rows, num_columns, subplot_idx)    
    plt.hist(angle_md[::stride,j], bins = 30, density = True, label = f"angle_md", alpha = .5)
    plt.hist(angle_im[::stride,j], bins = 30, density = True, label = f"angle_im", alpha = .5)
    plt.title('angle')
    plt.legend()
    subplot_idx += 1
    print(subplot_idx)    
    
for j in range(dihedral_md.shape[-1]):
    plt.subplot(num_rows, num_columns, subplot_idx)    
    plt.hist(dihedral_md[::stride,j], bins = 30, density = True, label = f"dihedral_md", alpha = .5)
    plt.hist(dihedral_im[::stride,j], bins = 30, density = True, label = f"dihedral_im", alpha = .5)
    plt.title('dihedral')
    plt.legend()
    subplot_idx += 1
    print(subplot_idx)    
    
plt.tight_layout()
plt.savefig(f"./output/{name}/bad_hist_im.pdf")

exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=NVT_IM_US
#SBATCH --time=03:00:00
#SBATCH --partition=xeon-p8
##SBATCH --ntasks=1
##SBATCH --cpus-per-task=2
##SBATCH --nodes=1
##SBATCH --array=0-39
##SBATCH --open-mode=truncate
##SBATCH --output=./slurm_output/NVT_IM_US_%a.txt


import numpy as np
import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import argparse
import pandas as pd
import mdtraj
from sys import exit
import os
import time
import pandas as pd
import os
import argparse
import pickle

parser = argparse.ArgumentParser()
parser.add_argument('--rank', type = int)
parser.add_argument('--size', type = int)
args = parser.parse_args()

rank = args.rank
size = args.size
print(f"rank: {rank}, size: {size}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                           index_col = 'name', comment = '#')
protein_names = protein_names.iloc[0:2, 0].tolist()
rmsd_centers_and_k = {}
for name in protein_names:
    with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
        data = pickle.load(file_handle)
        rmsd_centers_and_k[name] = data
sizes = [rmsd_centers_and_k[name]['size'] for name in protein_names]
cumsum_sizes = np.cumsum(sizes)

exit()

if rank >= cumsum_sizes[-1]:
    exit()

idx_name = np.searchsorted(cumsum_sizes, rank, side = 'right')
name = protein_names[idx_name]

if idx_name == 0:
    idx_window = rank
else:
    idx_window = rank - cumsum_sizes[idx_name - 1]    

rmsd_center = rmsd_centers_and_k[name]['centers'][idx_window]
rmsd_k = rmsd_centers_and_k[name]['k']

print("sizes:", sizes)        
print("cumsum_sizes:", cumsum_sizes)
print(f"rank: {rank}, name: {name}, idx_window: {idx_window}")
print(f"rmsd_k: {rmsd_k:.2f}, rmsd_center: {rmsd_center:.3f}")

with open(f"./output/{name}/system_im.xml", 'r') as file_handle:
    xml = file_handle.read()    
system = omm.XmlSerializer.deserialize(xml)

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)
xyz_init = traj_ref.xyz[0]

particle_index = list(range(psf.n_residues))
rmsd_force = omm.RMSDForce(traj_ref.xyz[0], particle_index)
cv_force = omm.CustomCVForce("0.5*rmsd_k*(rmsd - rmsd_0)^2")
cv_force.addCollectiveVariable('rmsd', rmsd_force)
cv_force.addGlobalParameter('rmsd_k', rmsd_k)
cv_force.addGlobalParameter('rmsd_0', rmsd_center)
system.addForce(cv_force)

## make an integrator
T = protein_info.loc[name, 'temperature']
fricCoef = 1./unit.picoseconds
stepsize = 5. * unit.femtoseconds
integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

platform = omm.Platform.getPlatformByName('Reference')
context = omm.Context(system, integrator, platform)
context.setPositions(xyz_init)

for k in np.linspace(0, rmsd_k, 1000):
    context.setParameter('rmsd_k', k)
    integrator.step(1000)    
    print(k)
    
context.setParameter('rmsd_k', rmsd_k)
integrator.step(10000)

os.makedirs(f"./output/{name}/traj_imus", exist_ok = True)
file_handle = open(f"./output/{name}/traj_imus/traj_{idx_window}.dcd", 'wb')
dcd_file = app.DCDFile(file_handle, psf.to_openmm(), dt = 200*unit.femtoseconds)

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)

start_time = time.time()
num_frames = len(traj_md)*10//sizes[idx_name] + 1

print(f"num_frames: {num_frames}")

for i in range(num_frames):
    integrator.step(1000)
    state = context.getState(getPositions = True)
    pos = state.getPositions()
    dcd_file.writeModel(pos)

    if (i + 1) % 100 == 0:
        print(i, flush = True)

print("time used: {:.2f}".format(time.time() - start_time))
file_handle.close()        



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=basis_lj
#SBATCH --time=1-00:00:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --mem=100G
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/basis_lj_%a.txt

import numpy as np
from scipy.interpolate import BSpline
from scipy.integrate import quad
import pickle
import mdtraj
from sys import exit
import os
import argparse
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from CLCG.utils.splines import *
from CLCG.utils.CL import *
import torch
torch.set_default_dtype(torch.double)
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
import pandas as pd
import ray
from scipy.sparse import csr_matrix

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = -1
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG_250K/{name}.dcd", psf, stride = 1)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)
stride = size // 10
traj_imus = traj_imus[::stride]

with open(f"./output/common/LJ_rmin.pkl", 'rb') as file_handle:
    r_min_dict = pickle.load(file_handle)
aa_pairs = list(r_min_dict.keys())
aa_pairs.sort()

num_aa_pairs = len(aa_pairs)
num_of_basis = 12

resnames = [res.name for res in psf.residues]

def sort_pair_name(pair):
    if pair[0] <= pair[1]:
        return pair
    else:
        tmp = pair[0]
        pair[0] = pair[1]
        pair[1] = tmp
        return pair

pair_indices = [(i,j) for i in range(psf.n_residues)
                for j in range(i+4, psf.n_residues)]

r_md = mdtraj.compute_distances(traj_md, pair_indices)
r_imus = mdtraj.compute_distances(traj_imus, pair_indices)

r_min = []
for i,j in pair_indices:
    pair = tuple(sort_pair_name([resnames[i], resnames[j]]))    
    r_min.append(r_min_dict[pair])
r_min= np.array(r_min)
r_max = 1.5

ray.init(ignore_reinit_error = True, _temp_dir = "/home/gridsan/dingxq/tmp/ray")

r_md = ray.put(r_md)
r_imus = ray.put(r_imus)
r_min = ray.put(r_min)

@ray.remote
def compute_lj_basis(j, r_md, r_imus, r_min):
    import sys
    sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
    from CLCG.utils.splines import bs_lj
    print(j)

    r = torch.from_numpy(np.copy(r_md[:,j]))
    basis_md = bs_lj(r, r_min[j], r_max, num_of_basis, omega = False)

    r = torch.from_numpy(np.copy(r_imus[:,j]))    
    basis_imus = bs_lj(r, r_min[j], r_max, num_of_basis, omega = False)

    return basis_md, basis_imus

res = ray.get([compute_lj_basis.remote(j, r_md, r_imus, r_min) for j in range(len(pair_indices))])

basis_md = torch.zeros((len(traj_md), num_of_basis*num_aa_pairs))
basis_imus = torch.zeros((len(traj_imus), num_of_basis*num_aa_pairs))

# basis_md = csr_matrix((len(traj_md), num_of_basis*num_aa_pairs))
# basis_imus = csr_matrix((len(traj_imus), num_of_basis*num_aa_pairs))

for k in range(len(pair_indices)):
    i,j = pair_indices[k]
    print(i,j, flush = True)
    pair = tuple(sort_pair_name([resnames[i], resnames[j]]))

    # for md 
    basis = res[k][0]
    idx = aa_pairs.index(pair)
    basis_md[:, idx*num_of_basis:(idx+1)*num_of_basis] += basis
    
    # data = basis.data
    # indices = basis.indices
    # indptr = basis.indptr
    # indices = indices + num_of_basis*aa_pairs.index(pair)
    # basis = csr_matrix((data, indices, indptr), shape = basis_md.shape)
    # basis_md = basis_md + basis

    ## for imus
    basis = res[k][1]
    basis_imus[:, idx*num_of_basis:(idx+1)*num_of_basis] += basis
    
    # data = basis.data
    # indices = basis.indices
    # indptr = basis.indptr
    # indices = indices + num_of_basis*aa_pairs.index(pair)
    # basis = csr_matrix((data, indices, indptr), shape = basis_imus.shape)
    # basis_imus = basis_imus + basis

# def scipy_csr_to_torch_coo(s):
#     s = s.tocoo()
#     s = torch.sparse_coo_tensor(np.array([s.row, s.col]),
#                                 s.data,
#                                 size = s.shape)
#     return s

# basis_md = scipy_csr_to_torch_coo(basis_md)
# basis_imus = scipy_csr_to_torch_coo(basis_imus)

torch.save({'md': basis_md,
            'imus': basis_imus},
           f"./output/{name}/basis_lj.pt")

exit()
if name != "CLN025":
    exit()
    
## for grid
basis_grid = {}
basis_grid['min'] = 0.1
basis_grid['max'] = 1.5
basis_grid['basis'] = []
basis_grid['omega'] = []

r_grid_min, r_grid_max = basis_grid['min'], basis_grid['max']
r_grid = torch.from_numpy(np.linspace(r_grid_min, r_grid_max, 1000))
r_grid = ray.put(r_grid)

@ray.remote
def compute_lj_basis_grid(r_grid, r_min):
    import sys
    sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
    from CLCG.utils.splines import bs_lj
    
    basis, omega = bs_lj(r_grid, r_min, r_max, num_of_basis, omega = True)
    return basis, omega

res = ray.get([compute_lj_basis_grid.remote(r_grid, r_min_dict[(aa1, aa2)]) for aa1, aa2 in aa_pairs])

basis_grid['basis'] = torch.stack([res[i][0] for i in range(len(res))])
basis_grid['omega'] = torch.stack([res[i][1] for i in range(len(res))])

torch.save(basis_grid, f"./output/common/basis_lj_grid.pt")
exit()



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=basis_rmsd
#SBATCH --time=00:30:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --mem=100G
#SBATCH --array=0-1
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/basis_rmsd_%a.txt

import numpy as np
import torch
torch.set_default_dtype(torch.float64)
import pandas as pd
from sys import exit
import simtk.unit as unit
from scipy import optimize
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pickle
import os
import argparse
import mdtraj
import simtk.unit as unit
import math
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from MMFlow import utils
from CLCG.utils.splines import *
from CLCG.utils.CL import *
from scipy.sparse import csr_matrix
import pandas as pd
import os
import itertools

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)

#job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
job_idx = 1
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")

traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf, stride = 1)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
stride = 10
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj = traj[::stride]
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)

traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

ss = pd.read_table("./info/secondary_structure.csv",
                   index_col = 'name', header = 0)
ss = eval(ss.loc[name, 'ss'])

if len(ss) == 0:
    exit()

basis = []
    
for idx_ss in range(len(ss)):
    particle_index = [list(range(start-1, end)) for start, end in ss[idx_ss]]
    particle_index = list(itertools.chain(*particle_index))

    rmsd_md = mdtraj.rmsd(traj_md, traj_ref, atom_indices = particle_index)
    rmsd_imus = mdtraj.rmsd(traj_imus, traj_ref, atom_indices = particle_index)

    rmsd_md = torch.from_numpy(rmsd_md).double()
    rmsd_imus = torch.from_numpy(rmsd_imus).double()

    rmsd_max = rmsd_md.max().item()
    num_of_basis = 15

    exit()
    
    basis_md = bs_rmsd(rmsd_md, rmsd_max, num_of_basis)
    basis_imus = bs_rmsd(rmsd_imus, rmsd_max, num_of_basis)

    rmsd_grid = torch.linspace(0, rmsd_max, steps = 1000)
    basis_grid = {}
    basis_grid['basis'] = bs_rmsd(rmsd_grid, rmsd_max, num_of_basis)
    basis_grid['min'] = 0.0
    basis_grid['max'] = rmsd_max

    basis.append({'md': basis_md,
                  'imus': basis_imus,
                  'grid': basis_grid})

torch.save(basis, f"./output/{name}/basis_rmsd.pt")
    
exit()    
    
    




import mdtraj
import argparse
import math
import numpy as np
import pickle
import matplotlib.pyplot as plt
import matplotlib as mpl
mpl.rcParams['font.size'] = 14
mpl.rcParams['axes.labelsize'] = 'large'
mpl.rcParams['xtick.labelsize'] = 'large'
mpl.rcParams['ytick.labelsize'] = 'large'
from matplotlib.backends.backend_pdf import PdfPages
from sys import exit
from itertools import product
import pandas as pd
import ray
import os

parser = argparse.ArgumentParser()
parser.add_argument('--elec_type', type = str)
parser.add_argument('--ss_type', type = str)
args = parser.parse_args()

elec_type = args.elec_type
ss_type = args.ss_type

#print(f"elec_type: {elec_type:10}, ss_type: {ss_type}")

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_names = protein_names.iloc[:, 0].tolist()
weight_decay_list = [5e-7, 1e-6, 2e-6]

weight_decay = 1e-6

protein_names_alt = {
    'CLN025': 'Chignolin',
    '2JOF': 'Trp-cage',
    '1FME': 'BBA',
    '2F4K': 'Villin',
    'GTT': 'WW domain',
    'NTL9': 'NTL9',
    '2WAV': 'BBL',
    'PRB': 'Protein B',
    'UVF': 'Homeodomain',
    'NuG2': 'Protein G',
    'A3D': '$\alpha$3D',
    'lambda': '$\lambda$-repressor'
}

with open(f"./output/plots/rmsd_and_rg.pkl", 'rb') as file_handle:
    data = pickle.load(file_handle)
    rmsd_md = data['rmsd_md']
    rg_md = data['rg_md']
    rmsd_cg = data['rmsd_cg']
    rg_cg = data['rg_cg']

for name in protein_names:
    rmsd_cg[weight_decay][name] = rmsd_cg[weight_decay][name][10000:]
    rg_cg[weight_decay][name] = rg_cg[weight_decay][name][10000:]    

#### rmsd ####
rmsd_max = {'CLN025': 0.80, '2JOF': 1.28, '1FME': 1.53, '2F4K': 1.53,
         'GTT': 1.53, 'NTL9': 1.30, '2WAV': 2.03, 'PRB': 2.33,
         'UVF': 2.03, 'NuG2': 1.63, 'A3D': 2.53, 'lambda': 1.73}

rg_max = {'CLN025': 0.93, '2JOF': 1.53, '1FME': 2.03, '2F4K': 1.83,
         'GTT': 1.53, 'NTL9': 1.23, '2WAV': 2.03, 'PRB': 2.33,
         'UVF': 2.03, 'NuG2': 1.63, 'A3D': 2.63, 'lambda': 1.73}
    
ncols = 4
nrows = 3

fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
fig.clf()
for name in protein_names:
    print(name)
    axes = fig.add_subplot(nrows, ncols, protein_names.index(name) + 1)
    bins = np.linspace(0, rmsd_max[name], 40)
    density, bin_edges = np.histogram(rmsd_md[name], bins = bins, density = True)
    F_md = -np.log(density)

    density, bin_edges = np.histogram(rmsd_cg[weight_decay][name], bins = bins, density = True)
    F_cg = -np.log(density)
    F_cg = F_cg - F_cg.min() + F_md.min()
    
    plt.plot((bins[0:-1]+bins[1:])/2., F_md, color = 'C1', label = 'All atom', linewidth = 3.0)
    plt.plot((bins[0:-1]+bins[1:])/2., F_cg, color = 'C0', label = 'CG', linewidth = 3.0)    
    axes.legend()
    if name == 'A3D':
        plt.xlabel("RMSD (nm)"
                   "\n"
                   r"$\alpha$-3D")
    elif name == 'lambda':
        plt.xlabel("RMSD (nm)"
                   "\n"
                   r"$\lambda$-repressor")
    else:
        xlabel = f'RMSD (nm) \n {protein_names_alt[name]}'
        plt.xlabel(xlabel)
        
    plt.ylabel(r'Free energy ($k_B$T)')    
    # axes.hist(rmsd_md[name], bins = bins, alpha = 0.5, label = 'All atom', density = True, color = 'C1', log = True)
    # axes.hist(rmsd_cg[weight_decay][name], bins = bins, alpha = 0.5, label = 'CG', density = True, color = 'C0', log = True)
    # if name == 'CLN025':
    #     axes.legend()
    # plt.xlim(-0.1, rmsd_max[name])

os.makedirs(f"./output/figures", exist_ok = True)
#plt.tight_layout()
plt.subplots_adjust(hspace = 0.5, wspace = 0.3)
plt.savefig(f"./output/figures/rmsd_free_energy.eps")


fig = plt.figure(figsize = (6.4*ncols, 4.8*nrows))
fig.clf()
for name in protein_names:
    print(name)
    axes = fig.add_subplot(nrows, ncols, protein_names.index(name) + 1)
    bins = np.linspace(0, rg_max[name], 40)
    density, bin_edges = np.histogram(rg_md[name], bins = bins, density = True)
    F_md = -np.log(density)

    density, bin_edges = np.histogram(rg_cg[weight_decay][name], bins = bins, density = True)
    F_cg = -np.log(density)
    F_cg = F_cg - F_cg.min() + F_md.min()
    
    plt.plot((bins[0:-1]+bins[1:])/2., F_md, color = 'C1', label = 'All atom', linewidth = 3.0)
    plt.plot((bins[0:-1]+bins[1:])/2., F_cg, color = 'C0', label = 'CG', linewidth = 3.0)    
    axes.legend()
    if name == 'A3D':
        plt.xlabel("Rg (nm)"
                   "\n"
                   r"$\alpha$-3D")
    elif name == 'lambda':
        plt.xlabel("Rg (nm)"
                   "\n"
                   r"$\lambda$-repressor")
    else:
        xlabel = f'Rg (nm) \n {protein_names_alt[name]}'
        plt.xlabel(xlabel)
        
    plt.ylabel(r'Free energy ($k_B$T)')    
    # axes.hist(rg_md[name], bins = bins, alpha = 0.5, label = 'All atom', density = True, color = 'C1', log = True)
    # axes.hist(rg_cg[weight_decay][name], bins = bins, alpha = 0.5, label = 'CG', density = True, color = 'C0', log = True)
    # if name == 'CLN025':
    #     axes.legend()
    # plt.xlim(-0.1, rg_max[name])

os.makedirs(f"./output/figures", exist_ok = True)
#plt.tight_layout()
plt.subplots_adjust(hspace = 0.5, wspace = 0.3)
plt.savefig(f"./output/figures/rg_free_energy.eps")

exit()



import mdtraj
import pandas as pd
import numpy as np
from sys import exit
import os
protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
protein_names = protein_names.iloc[:, 0].tolist()

N = 250_000
os.makedirs(f"./data/traj_CG_250K", exist_ok = True)
for name in protein_names:
    psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")    
    traj = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf)
    stride = len(traj) // N
    traj = traj[::stride]
    traj.save_dcd(f"./data/traj_CG_250K/{name}.dcd")
    print(f"{name:10}: {len(traj):10,}")




#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=make_system
#SBATCH --time=01:00:00
#SBATCH --partition=xeon-p8
#SBATCH --ntasks=1
#SBATCH --nodes=1
#SBATCH --array=0-11
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/make_system_%a.txt

import simtk.openmm.app  as app
import simtk.openmm as omm
import simtk.unit as unit
import pickle
import math
from sys import exit
from collections import defaultdict
import numpy as np
import mdtraj
from collections import defaultdict
import pandas as pd
import argparse
from scipy.interpolate import BSpline
import os
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/MMFlow")
from MMFlow import utils
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG/")
from CLCG.utils.make_system import *
import itertools

protein_names = pd.read_csv("./info/protein_names.txt",
                            comment = "#",
                            header = None)

# weight_decay_list = [1e-8,
#                      1e-7, 2e-7, 5e-7,
#                      1e-6, 2e-6, 5e-6,
#                      1e-5]

weight_decay_list = [5e-7,
                     1e-6, 2e-6]


job_idx = int(os.environ['SLURM_ARRAY_TASK_ID'])
name = protein_names.iloc[job_idx, 0]

print(f"name: {name}", flush = True)

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
resnames = [residue.name for residue in psf.residues]
for i in range(len(resnames)):
    if resnames[i] == "NLE":
        resnames[i] = "ILE"
nres = len(resnames)

protein_info = pd.read_csv("./info/protein_temperature_and_ionic_strength.txt",
                           index_col = 'name',
                           comment = '#')    
T = protein_info.loc[name, 'temperature']
ionic_strength = protein_info.loc[name, 'ionic_strength']

kbT = unit.BOLTZMANN_CONSTANT_kB * T *unit.kelvin * unit.AVOGADRO_CONSTANT_NA
kbT_kJ_per_mole = kbT.value_in_unit(unit.kilojoule_per_mole)
beta = 1/kbT_kJ_per_mole

Kb = unit.BOLTZMANN_CONSTANT_kB*unit.AVOGADRO_CONSTANT_NA
Kb = Kb.value_in_unit(unit.kilojoule_per_mole/unit.kelvin)

with open(f"./output/{name}/coor_transformer.pkl", 'rb') as file_handle:
    coor_transformer = pickle.load(file_handle)

AA_info = pd.read_csv("./info/amino_acids_with_learned_sigmas.csv",
                      index_col = 'name')
masses = [AA_info.loc[r.name, 'mass'] for r in psf.residues]


def make_custom_elec_force(elec_type):
    if elec_type == 'simple':
        formula = ["ONE_4PI_EPS0*charge1*charge2/(r*epsilon)",
                   "ONE_4PI_EPS0 = 138.935456",
                   "epsilon = 78.4"
        ]
    elif elec_type == 'fshift':
        formula = ["ONE_4PI_EPS0*charge1*charge2/epsilon*fr*step(rc - r)",
                   "fr = 1/r - 5/(3*rc) + 5*r^3/(3*rc^4) - r^4/rc^5",              
                   "ONE_4PI_EPS0 = 138.935456",
                   "epsilon = 15",
                   "rc = 1.2"
        ]
    elif elec_type == 'DH':
        formula = ["ONE_4PI_EPS0*charge1*charge2/(r*eps)*exp(-r/lambda_D)*step(r-r_cut)",
                   "lambda_D = sqrt(eps*constant*T/ionic_strength)",
                   "constant = 2.3811197070710097 / 10^6",
                   "eps = A + B/(1 + kappa*exp(-lambda*B*r))",
                   "B = 78.4 - A; A = -8.5525; kappa = 7.7839; lambda = 0.03627",
                   "ONE_4PI_EPS0 = 138.935456",
                   "r_cut = 0.326"
        ]
    elif elec_type == 'DH_2':
        formula = ["ONE_4PI_EPS0*charge1*charge2/(r*eps)*exp(-r/lambda_D)",
                   "lambda_D = sqrt(eps*constant*T/ionic_strength)",
                   "constant = 2.3811197070710097 / 10^6",
                   "eps = 78.4",
                   "ONE_4PI_EPS0 = 138.935456",
        ]
    else:
        raise ValueError("elec_type has to be either simple, fshift, DH, or DH_2")
    
    custom_elec_force = omm.CustomNonbondedForce(";".join(formula))
    custom_elec_force.addPerParticleParameter('charge')

    if elec_type == 'DH' or 'DH_2':    
        custom_elec_force.addGlobalParameter("T", T)
        custom_elec_force.addGlobalParameter("ionic_strength", ionic_strength)
    
    return custom_elec_force

def sort_pair_name(pair):
    if pair[0] <= pair[1]:
        return pair
    else:
        tmp = pair[0]
        pair[0] = pair[1]
        pair[1] = tmp
        return pair

for weight_decay in weight_decay_list:
    #for elec_type, ss_type in itertools.product(['simple', 'fshift'], ['simple', 'extended']):
    #for elec_type, ss_type in itertools.product(['DH'], ['simple', 'extended']):
    for elec_type, ss_type in itertools.product(['DH_2'], ['simple']):                
        print(f"weight_decay: {weight_decay:.3E}, elec_type: {elec_type:8}, ss_type: {ss_type:8}")
        data = torch.load(f"./output/{name}/FF/FF_elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.pt",
                          map_location=torch.device('cpu'))
        FF = data['FF']
        bonded_ff, lj_ff, rmsd_ff = FF['bonded'], FF['lj'], FF['rmsd']

        ## bonded terms
        system = make_system(masses,
                             coor_transformer,
                             bonded_ff,
                             protein_info.loc[name, 'temperature'],
                             jacobian = False)

        ## electrostatic
        custom_elec_force = make_custom_elec_force(elec_type)
        custom_elec_force.setForceGroup(1)
        for i in range(psf.n_residues):
            resname = psf.residue(i).name
            charge = AA_info.loc[resname, 'charge']
            custom_elec_force.addParticle([charge])
        bonds = [[i, i+1] for i in range(psf.n_residues - 1)]
        custom_elec_force.createExclusionsFromBonds(bonds, 3)    
        system.addForce(custom_elec_force)        

        ## LJ
        with open(f"./output/common/LJ_rmin.pkl", 'rb') as file_handle:
            r_min_dict = pickle.load(file_handle)
        aa_pairs = list(r_min_dict.keys())
        aa_pairs.sort()

        ulj = lj_ff['U'].numpy()
        lj_min = lj_ff['min']
        lj_max = lj_ff['max']    
        func = omm.Continuous2DFunction(xsize = ulj.shape[0],
                                        ysize = ulj.shape[1],
                                        values = ulj.reshape(-1, order = 'F'),
                                        xmin = 0., xmax = float(ulj.shape[0] - 1),
                                        ymin = lj_min, ymax = lj_max)

        LJ_force = omm.CustomCompoundBondForce(2, f"U_LJ(idx, r); r = distance(p1,p2)")
        LJ_force.addTabulatedFunction(f"U_LJ", func)
        LJ_force.addPerBondParameter('idx')

        for i in range(len(resnames)):
            for j in range(i+4, len(resnames)):
                aa1, aa2 = resnames[i], resnames[j]
                pair = tuple(sort_pair_name([aa1, aa2]))
                idx = float(aa_pairs.index(pair))
                LJ_force.addBond([i,j], [idx])
        LJ_force.setForceGroup(2)        
        system.addForce(LJ_force)

        ## rmsd
        ss = pd.read_table(f"./info/secondary_structure_{ss_type}.csv",
                           index_col = 'name', header = 0)
        ss = eval(ss.loc[name, 'ss'])
        traj_ref = mdtraj.load_xyz(f"./output/{name}/reference_structure.xyz", psf)

        rmsd_min = rmsd_ff['min']
        rmsd_max = rmsd_ff['max']
        rmsd_U = rmsd_ff['U']

        for idx_ss in range(len(ss)):
            particle_index = [list(range(start-1, end)) for start, end in ss[idx_ss]]
            particle_index = list(itertools.chain(*particle_index))

            rmsd_force = omm.RMSDForce(traj_ref.xyz[0], particle_index)
            cv_force = omm.CustomCVForce("Kb*T*u_rmsd(rmsd)")
            cv_force.addCollectiveVariable('rmsd', rmsd_force)
            u_rmsd = omm.Continuous1DFunction(rmsd_U[idx_ss].numpy(),
                                              rmsd_min[idx_ss].item(),
                                              rmsd_max[idx_ss].item())
            cv_force.addTabulatedFunction('u_rmsd', u_rmsd)
            cv_force.addGlobalParameter('Kb', Kb)
            cv_force.addGlobalParameter('T', T)
            system.addForce(cv_force)

        xml = omm.XmlSerializer.serialize(system)
        os.makedirs(f"./output/{name}/full_system", exist_ok = True)
        with open(f"./output/{name}/full_system/elec_type_{elec_type}_ss_type_{ss_type}_weight_decay_{weight_decay:.3E}.xml", 'w') as file_handle:
            file_handle.write(xml)

        fricCoef = 1./unit.picoseconds
        stepsize = 2. * unit.femtoseconds
        integrator = omm.LangevinMiddleIntegrator(T, fricCoef, stepsize)

        platform = omm.Platform.getPlatformByName('Reference')
        context = omm.Context(system, integrator, platform)

        context.setPositions(traj_ref.xyz[0])
        state = context.getState(getEnergy = True)
        potential_energy = state.getPotentialEnergy()
        print(potential_energy)
    
exit()



import matplotlib.pyplot as plt
import numpy as np

loss_record = []
with open("./slurm_output/CL.txt", 'r') as file:
    flag = 0
    for line in file.readlines():
        line = line.strip()
        if 'X0' in line:
            flag += 1
        if flag == 2:
            if line[0:4] == "At i":
                loss = line[23:34].replace('D', 'E')
                loss_record.append(float(loss))

loss = np.array(loss_record)

fig = plt.figure()
fig.clf()
plt.plot(range(1, len(loss)+1), loss)
plt.xscale('log')
plt.xlabel('Number of iterations')
plt.ylabel('Aggregated loss with weight decay')
plt.savefig('./output/figures/loss.eps')



#!/home/gridsan/dingxq/.conda/envs/openmm_torch/bin/python

# Created at 2021/05/18 15:54:25

#SBATCH --job-name=basis_lj
#SBATCH --time=00:30:00
#SBATCH --partition=xeon-p8
#SBATCH --nodes=1
#SBATCH --exclusive
#SBATCH --mem=100G
#SBATCH --array=0-1
#SBATCH --open-mode=truncate
#SBATCH --output=./slurm_output/basis_lj_%a.txt

import numpy as np
from scipy.interpolate import BSpline
from scipy.integrate import quad
import pickle
import mdtraj
from sys import exit
import os
import argparse
import sys
sys.path.append("/home/gridsan/dingxq/my_package_on_github/CLCG")
from CLCG.utils.splines import *
from CLCG.utils.CL import *
import torch
torch.set_default_dtype(torch.double)
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
mpl.rc('font', size = 16)
mpl.rc('axes', titlesize = 'large', labelsize = 'large')
mpl.rc('xtick', labelsize = 'large')
mpl.rc('ytick', labelsize = 'large')
import pandas as pd
import ray

protein_names = pd.read_csv("./info/protein_names.txt", comment = "#", header = None)
job_idx = 1
name = protein_names.iloc[job_idx, 0]

psf = mdtraj.load_psf(f"./data/structures/{name}/{name}.psf")
traj_md = mdtraj.load_dcd(f"./data/traj_CG/{name}.dcd", psf, stride = 1)

with open(f"./output/{name}/rmsd_centers_and_k_for_imus.pkl", 'rb') as file_handle:
    rmsd_centers_and_k = pickle.load(file_handle)
size = rmsd_centers_and_k['size']
traj_imus = []
stride = 10
for rank in range(size):
    print(f"rank: {rank}")
    traj = mdtraj.load_dcd(f"./output/{name}/traj_imus/traj_{rank}.dcd", psf)
    traj = traj[::stride]
    traj_imus.append(traj)
traj_imus = mdtraj.join(traj_imus)

indices = [(i,j) for i in range(psf.n_residues) for j in range(i+4, psf.n_residues)]
r_md = mdtraj.compute_distances(traj_md, indices)
r_imus = mdtraj.compute_distances(traj_imus, indices)

ncol = 6
nrow = r_md.shape[-1]//ncol + 1

fig = plt.figure(figsize = (6.4*ncol, 4.8*nrow))
fig.clf()
for i in range(r_md.shape[1]):
    plt.subplot(nrow, ncol, i+1)
    plt.hist(r_md[:, j], bins = 100, density = True, alpha = 0.5, label = 'All atom', log = True)
    plt.hist(r_imus[:, j], bins = 100, density = True, alpha = 0.5, label = 'IMUS', log = True)
    plt.tight_layout()

os.makedirs(f"./output/LJ/{name}/CG_simulations/", exist_ok = True)
fig.savefig(f"./output/LJ/{name}/CG_simulations/pairwise_distances_hist_{noise_model}_weight_decay_{weight_decay:.3E}.pdf")
